import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

from pptx import Presentation

import rwe.clients.aou as aou
from rwe.generate_report import generate_rwe_report
from tests._fixtures import write_azn_binary_csv, write_azn_continuous_csv, write_genebass_csv


class PackageSmokeTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.output_dir = Path(self.tmpdir.name)
        self.gene = "TESTGENE"
        write_genebass_csv(self.output_dir, self.gene)
        write_azn_binary_csv(self.output_dir, self.gene)
        write_azn_continuous_csv(self.output_dir, self.gene)

    def tearDown(self) -> None:
        self.tmpdir.cleanup()

    def test_generate_rwe_report_ukb_only_smoke(self) -> None:
        def _unexpected_aou_call(*args, **kwargs):
            raise AssertionError("AoU path should not be used in UKB-only smoke test")

        out_docx = self.output_dir / "RWE_TESTGENE_report.docx"
        with patch.object(aou, "generate_aou_variant_info_demographics_report", side_effect=_unexpected_aou_call), \
            patch.object(aou, "generate_aou_clinical_report", side_effect=_unexpected_aou_call), \
            patch.object(aou, "generate_aou_labs_measurements_report", side_effect=_unexpected_aou_call), \
            patch.object(aou, "generate_aou_survey_report", side_effect=_unexpected_aou_call), \
            patch.object(aou, "generate_aou_homozygous_lof_carriers_report", side_effect=_unexpected_aou_call), \
            patch.object(aou, "generate_aou_indication_specific_report", side_effect=_unexpected_aou_call), \
            patch("rwe.clients.genebass.get_gene_info", return_value=(self.gene, "ENSG_TEST", "1")), \
            patch("rwe.utils.presentation.gnomad.get_gnomad_constraint_info", return_value={
                "input_gene": self.gene,
                "gene": self.gene,
                "loeuf": 0.42,
                "nhetz": 98,
                "nhomz": 2,
            }), \
            patch("rwe.utils.presentation.gbs.get_genebass_carrier_info", return_value={
                "gene": self.gene,
                "nhetz": 40,
                "nhomz": 1,
            }), \
            patch("rwe.utils.presentation.aou.get_aou_carrier_info", side_effect=_unexpected_aou_call):
            result = generate_rwe_report(
                gene=self.gene,
                chrm="1",
                out_docx_path=str(out_docx),
                indications="obesity",
                allofus=False,
                genebass=True,
                astrazeneca=True,
                generate_pptx=True,
            )

        self.assertTrue(Path(result["docx_path"]).exists())
        self.assertTrue(Path(result["pptx_path"]).exists())

        prs = Presentation(result["pptx_path"])
        slide4_text = "\n".join(shape.text for shape in prs.slides[3].shapes if getattr(shape, "has_text_frame", False))
        self.assertIn("LOEUF score: 0.420.", slide4_text)
        self.assertIn("gnomAD and UK Biobank cohorts have 138 (98 + 40) heterozygous and 3 (2 + 1) homozygous", slide4_text)


if __name__ == "__main__":
    unittest.main()



import os
import re
from functools import lru_cache
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing as mp
from scipy import stats
from docx import Document
from docx.shared import Inches
import matplotlib.pyplot as plt
from importlib.resources import as_file, files

import rwe.utils.helpers as uth
from rwe.plots.variant_info import create_variant_frequency_plots
from rwe.plots.demographics import demographics_plot
from rwe.plots.clinical import manhattan
from rwe.plots.measurements import plot_measurements
from rwe.plots.surveys import plot_survey_questions

_AOU_PTV_COUNTS_RESOURCE = "nsamples_ptv_gene.csv.gz"

############### Variant Info and Demographics ###############
def keep_most_del_plof(consequences):
    consequences =  set([c.strip() for c in consequences.split(",")])
    rank = {
        "frameshift_variant": 0,
        "stop_gained": 1,
        "splice_acceptor_variant": 2,
        "splice_donor_variant": 3,
        "start_lost": 4,
        "stop_lost": 5
    }
    return min(consequences, key=lambda x: rank.get(x, float('inf')))

def get_individual_info(bucket, chrm, gene, zygosity="hetz"):
    # TODO: need to change bucket folder based on biobanking updates
    person_df = pd.read_csv(f"{bucket}/data/rwe_info/raw/person.csv.gz")
    # TODO: need to change bucket folder based on biobanking updates
    variant_df = pd.read_csv(f"{bucket}/data/rwe_info/genes/chr{chrm}/{gene}/lof_{zygosity}.csv.gz")
    variant_df["consequence"] = variant_df.consequence.apply(keep_most_del_plof)
    person_df["cases"] = person_df.person_id.isin(variant_df[f"{zygosity}_samples"])
    person_df["group"] = np.where(person_df["cases"], "Cases", "Controls") 
    return person_df, variant_df

def generate_aou_variant_info_demographics_report(doc: Document, chrm: str, gene: str, zygosity: str ="hetz") -> Document:
    from rwe.parsers.aou.config import BUCKET
    person_df, variant_df = get_individual_info(BUCKET, chrm, gene, zygosity)
    if person_df.empty or variant_df.empty:
        doc.add_paragraph("No variant or individual level data provided to generate figures.")
        doc.add_page_break()
        return doc
    # 1) Variant information figure
    doc.add_paragraph()  # spacing
    doc.add_heading("Variant level information in All of Us cohort", level=2)
    fig1, ax_bar, top_show, bottom_show = create_variant_frequency_plots(variant_df)
    fig1_path = uth._save_fig_to_tmp(fig1, basename="variant_information", dpi=300)
    plt.close(fig1)  # close the figure to free memory
    doc.add_paragraph()  # spacing
    doc.add_picture(fig1_path, width=Inches(6.5))
    uth._add_figure_caption(doc, "Variant carrier counts by consequence.")
    doc.add_paragraph()  # spacing
    # Add tables for most/least frequent variants
    top_show.columns = ['Locus', 'Alleles', 'Consequence', 'Carriers (N)']
    bottom_show.columns = ['Locus', 'Alleles', 'Consequence', 'Carriers (N)']
    uth._add_table_to_doc(doc, top_show, uth._add_table_title(doc, f"Top 10 most frequent pLoF variants for {gene} in AoU"),
                    ['Locus', 'Alleles', 'Consequence', 'Carriers (N)'], list(map(Inches, [1.35, 2, 1.85, 0.8]))
                    )
    uth._add_table_to_doc(doc, bottom_show, uth._add_table_title(doc, f"Top 10 least frequent pLoF variants for {gene} in AoU"),
                    ['Locus', 'Alleles', 'Consequence', 'Carriers (N)'], list(map(Inches, [1.35, 2, 1.85, 0.8]))
                    )
    doc.add_page_break() 
    # 2) Demographics figure
    doc.add_paragraph()  # spacing
    doc.add_heading("Demographics information in All of Us cohort", level=2)
    fig2, _axes2 = demographics_plot(person_df)
    fig2_path = uth._save_fig_to_tmp(fig2, basename="demographics", dpi=300)
    plt.close(fig2)  # close the figure to free memory
    doc.add_paragraph()  # spacing
    doc.add_picture(fig2_path, width=Inches(4))
    uth._add_figure_caption(doc, "Demographics of cases vs controls (age, ancestry, sex at birth, ethnicity).")
    doc.add_page_break() 
    return doc


############### Clinical Records ###############
def clean_aou_phewas(phewas_file, tool="regenie", version="1.2"):
    if tool=="phetk":
        if version == "1.2":
            df = pd.read_csv(phewas_file)
            df = df.loc[df.converged==True]
        elif version =="X":
            df = pd.read_csv(phewas_file, sep="\t")
            df = df.loc[(df.ancestry == "all")&(df.converged==True)]
        else:
            raise ValueError(f"Unsupported AoU PheWAS version: {version}")
    elif tool=="regenie":
        df = pd.read_csv(phewas_file)
        # keep plof strict mask only, aaf bin remove singletons, then merge by the most significant aaf_bin
        df = df.loc[(df["mask"]=="pLoF_strict")&(df.aaf_bin!="singleton")].copy()
        df = df.loc[df.groupby("pheno")["pvalue"].idxmin()].sort_values("pvalue")
        # from assets add the category of each phecode
        labels_resource = files("rwe").joinpath("assets/phecodeX_R_labels.csv")
        with as_file(labels_resource) as labels_path:
            labels_df = pd.read_csv(labels_path, usecols=["phenotype", "group"], dtype={"phenotype": str})
        labels_df = labels_df.drop_duplicates(subset=["phenotype"])
        pheno_to_group = labels_df.set_index("phenotype")["group"]
        df["category"] = df["pheno"].astype(str).map(pheno_to_group)
    else:
        raise ValueError(f"Unsupported AoU PheWAS tool: {tool}")
    return df

def get_aou_manhattan(df, gene):
    name_col = "description"
    p_col    = "pvalue"
    beta_col = "beta"
    cat_col  = "category"
    sig_p = 0.05/len(df)
    fig, ax, plot_df = manhattan(
        df,
        p_col=p_col,
        category_col=cat_col,
        label_col=name_col,
        beta_col=beta_col,
        sig_p=sig_p,
        top_k_labels=5,
        title=f"{gene} PheWAS AoU",
    )
    return fig, plot_df

def generate_aou_clinical_report(doc, chrm, gene, zygosity, output_dir):
    from rwe.parsers.aou.config import BUCKET
    phewas_file = f"{BUCKET}/data/associations/results/burden/plof/aou_phecodeX/chr{chrm}/{gene}.csv.gz"
    doc.add_heading("Phenome-wide Association Study (PheWAS) in All of Us cohort", level=2)
    if uth._gcs_size(phewas_file, BUCKET) > 0: 
        df = clean_aou_phewas(phewas_file)
        # Save table to output dir
        df.to_csv(os.path.join(output_dir, f"{gene}_aou_phewas.csv"), index=False)
        fig, plot_df = get_aou_manhattan(df, gene)
        fig_path = uth._save_fig_to_tmp(fig, basename="aou_phewas", dpi=300)
        plt.close(fig)  # close the figure to free memory
        doc.add_paragraph()  # spacing
        doc.add_picture(fig_path, width=Inches(6.5))
        uth._add_figure_caption(doc, f"{gene} PheWAS results from All of Us.")
        doc.add_paragraph()  # spacing

        # Table: Top 10 significant hits (sorted by p-value)
        top_significant = plot_df.nsmallest(10, 'pvalue')
        columns = ['description', 'category', 'beta', 'pvalue']
        # Rename columns for display
        display_df = top_significant[columns].copy()
        display_df.columns = ['Phecode', 'Category', 'Effect', 'P-value']
        uth._add_table_to_doc(doc, display_df, uth._add_table_title(doc, f"Top 10 significant associations for {gene} in AoU"),
                        ['Phecode', 'Category', 'Effect', 'P-value'], list(map(Inches, [2.75, 1.75, 0.75, 0.75]))
                        )

        # Table: Top 10 negative beta hits (beta < 0, sorted by p-value)
        negative_beta = plot_df[plot_df['beta'] < 0].nsmallest(10, 'pvalue')
        if len(negative_beta) > 0:
            display_df_neg = negative_beta[columns].copy()
            display_df_neg.columns = ['Phecode', 'Category', 'Effect', 'P-value']
            uth._add_table_to_doc(doc, display_df_neg, uth._add_table_title(doc, f"Top 10 protective associations for {gene} in AoU"),
                            ['Phecode', 'Category', 'Effect', 'P-value'], list(map(Inches, [2.75, 1.75, 0.75, 0.75]))
                            )

        doc.add_page_break()  # spacing
    else:
        doc.add_paragraph(f"No AoU PheWAS results found for {gene}.")
        doc.add_page_break()  # spacing
    return doc

def generate_aou_indication_specific_report(doc, chrm, gene, output_dir, keywords):
    doc.add_heading("Indication specific PheWAS results in All of Us cohort", level=2)
    doc.add_paragraph("The keywords searched to generate the report were: " + ", ".join(keywords))
    if not os.path.exists(os.path.join(output_dir, f"{gene}_aou_phewas.csv")):
        doc.add_paragraph("No PheWAS results available to generate indication-specific report.")
        doc.add_page_break()
        return doc
    # TODO: update phewas results to change indication specific report
    phewas_df = pd.read_csv(os.path.join(output_dir, f"{gene}_aou_phewas.csv"))
    escaped_keywords = [re.escape(k) for k in keywords if k]
    if not escaped_keywords:
        doc.add_paragraph("No valid indication keywords were provided.")
        doc.add_page_break()
        return doc
    keyword_pattern = "|".join(escaped_keywords)
    phewas_df = phewas_df.loc[phewas_df.description.str.contains(keyword_pattern, case=False, na=False)]
    columns = ['description', 'category', 'beta', 'pvalue']
    display_df = phewas_df[columns].copy().sort_values("pvalue").reset_index(drop=True)
    if len(display_df) == 0:
        doc.add_paragraph("No associations found for the specified indications. Please broaden the indication keywords and try again.")
        doc.add_page_break()
    else:
        display_df.columns = ['Phecode', 'Category', 'Effect', 'P-value']
        uth._add_table_to_doc(doc, display_df, uth._add_table_title(doc, f"Indication-specific associations for {gene} in AoU"),
                        ['Phecode', 'Category', 'Effect', 'P-value'], list(map(Inches, [2.75, 1.75, 0.75, 0.75]))
                        )
        doc.add_page_break()
    return doc

############### Labs and Measurements ###############
def compare(df, measurement):
    mdf = df.loc[df.measurement==measurement].copy()
    ctrls = mdf.loc[mdf.cases==False].drop_duplicates(["person_id", "measurement_concept_id"])
    cases = mdf.loc[mdf.cases==True].drop_duplicates(["person_id", "measurement_concept_id"])
    x = pd.to_numeric(cases["median_value"], errors="coerce").dropna().to_numpy()
    y = pd.to_numeric(ctrls["median_value"], errors="coerce").dropna().to_numpy()
    caq1, caq2, caq3 = pd.Series(x).quantile([0.25, 0.5, 0.75])
    ctq1, ctq2, ctq3 = pd.Series(y).quantile([0.25, 0.5, 0.75])
    # guard against empty arrays
    if x.size<20 or y.size<20:
        return (measurement, np.nan, np.nan, len(cases), len(ctrls), np.nan, np.nan, np.nan, np.nan, np.nan, np.nan)
    u_stat, p_mwu = stats.mannwhitneyu(x, y, alternative="two-sided")
    return (measurement, float(u_stat), float(p_mwu), int(len(cases)), int(len(ctrls)), caq1, caq2, caq3, ctq1, ctq2, ctq3)

_G_DF = None

def _init_worker(df):
    global _G_DF
    _G_DF = df

def _worker_compare(measurement):
    # uses global df set once per process
    return compare(_G_DF, measurement)

def run_parallel(df, measurements=None, n_jobs=None, chunksize=10):
    if measurements is None:
        measurements = sorted(df["measurement"].dropna().unique().tolist())
    if n_jobs is None:
        n_jobs = max(1, mp.cpu_count() - 1)
    with mp.Pool(processes=n_jobs, initializer=_init_worker, initargs=(df,)) as pool:
        rows = list(
            tqdm(
                pool.imap_unordered(_worker_compare, measurements, chunksize=chunksize),
                total=len(measurements),
                desc="Processing measurements"
            )
        )
    results_df = pd.DataFrame(
        rows, columns=["measurement", "u_stat", "p_mwu", "n_cases", "n_ctrls", "q1_case", "median_case", "q3_case", "q1_ctrl", "median_ctrl", "q3_ctrl"]
    )
    return results_df

def compare_aou_labs_measurements_raw(numerical_measurements_df, measurements=None):
    res_df = run_parallel(numerical_measurements_df, measurements=measurements, n_jobs=None, chunksize=20)
    all_lab = res_df.sort_values("p_mwu").loc[:, ["measurement", "p_mwu", "u_stat", "n_cases", "n_ctrls"]]
    all_lab.columns = ["Measurement", "P-value", "U statistic", "N cases", "N controls"]
    return res_df, all_lab

def get_aou_labs_measurements_burden(chrm, gene):
    from rwe.parsers.aou.config import BUCKET, MEASUREMENT_NAMES_MAP
    res_df = pd.read_csv(f"{BUCKET}/data/associations/results/burden/plof/measurements/chrm{chrm}/{gene}.csv.gz")
    mn_map = {v:k for k,v in MEASUREMENT_NAMES_MAP.items()}
    res_df["pheno"] = res_df["pheno"].map(mn_map).fillna(res_df["pheno"])
    res_df = res_df.loc[res_df["mask"]!="singleton"]
    res_df = res_df.loc[res_df.groupby("pheno")["pvalue"].idxmin()]
    all_lab = res_df.sort_values("pvalue").loc[:, ["pheno", "pvalue", "beta", "se", "n"]]
    all_lab.columns = ["Measurement", "P-value", "Beta", "SE", "N"]
    return res_df, all_lab

def generate_aou_labs_measurements_report(doc: Document, chrm: str, gene: str, zygosity: str) -> Document:
    from rwe.parsers.aou.config import MEASUREMENT_GROUPS, BUCKET
    numerical_measurements_df = pd.read_parquet(f"{BUCKET}/data/rwe_info/processed/selected_numerical_measurements.parquet")
    variant_df = pd.read_csv(f"{BUCKET}/data/rwe_info/genes/chr{chrm}/{gene}/lof_{zygosity}.csv.gz")
    numerical_measurements_df["cases"] = numerical_measurements_df.person_id.isin(variant_df[f"{zygosity}_samples"])
    measurements = numerical_measurements_df["measurement"].unique().tolist()
    # Burden tests from REGENIE
    res_df, all_lab = get_aou_labs_measurements_burden(chrm, gene)
    doc.add_heading("Association study results of labs and measurements in All of Us cohort", level=2)
    doc.add_paragraph()  # spacing
    uth._add_table_to_doc(
        doc, all_lab, uth._add_table_title(
            doc, f"Lab measurements for {gene} pLoF carriers in AoU"), 
            all_lab.columns.tolist(), list(map(Inches, [3.5, 0.75, 0.5, 0.5, 0.75]))
    )
    doc.add_page_break()  # spacing
    for k,v in MEASUREMENT_GROUPS.items():
        f, a = plot_measurements(
            numerical_measurements_df, v, col="median_value", 
            res_df=res_df, res_name_col="pheno", res_p_col="pvalue",
        )
        fig_path = uth._save_fig_to_tmp(f, basename=f"aou_measurements_{k}", dpi=300)
        plt.close(f)  # close the figure to free memory
        doc.add_paragraph()  # spacing
        doc.add_heading(f"{k.capitalize()} profile of {gene} pLoF carriers in All of Us cohort", level=2)
        doc.add_picture(fig_path, width=Inches(6.5))
        uth._add_figure_caption(doc, f"{gene} pLoF carrier {k} results from All of Us.")
        doc.add_page_break()  # spacing
    return doc

############### Surveys ###############
def clean_aou_surveys(df, survey_col="survey", question_col="question", answer_col="answer_category", zygosity="hetz"):
    df = df.copy()
    df[question_col] = df[question_col].str.replace(r"^.*?:\s*", "", regex=True, n=1)
    df[answer_col] = df[answer_col].str.replace(r"^.*?:\s*", "", regex=True, n=1)
    return df

def generate_aou_survey_report(doc: Document, chrm: str, gene: str, zygosity: str) -> Document:
    from rwe.parsers.aou.config import BUCKET
    survey_df = pd.read_parquet(f"{BUCKET}/data/rwe_info/processed/selected_surveys.parquet")
    variant_df = pd.read_csv(f"{BUCKET}/data/rwe_info/genes/chr{chrm}/{gene}/lof_{zygosity}.csv.gz")
    survey_df["cases"] = survey_df.person_id.isin(variant_df[f"{zygosity}_samples"])
    survey_df = clean_aou_surveys(survey_df, zygosity=zygosity)
    fig, ax = plot_survey_questions(survey_df)
    fig_path = uth._save_fig_to_tmp(fig, basename="aou_surveys", dpi=300)
    plt.close(fig)  # close the figure to free memory
    doc.add_paragraph()  # spacing
    doc.add_picture(fig_path, width=Inches(6.5))
    uth._add_figure_caption(doc, f"{gene} pLoF carrier survey results from All of Us.")
    doc.add_paragraph()  # spacing
    doc.add_page_break()
    return doc

############### Homozygous pLoF carriers ###############
def generate_aou_homozygous_lof_carriers_report_summary(doc, chrm, gene):
    from rwe.parsers.aou.config import BUCKET
    pdf, vdf = get_individual_info(BUCKET, chrm, gene, zygosity="homo")
    samples = set(pdf.loc[pdf.cases==True, "person_id"].values)
    # Demographics table for homozygous carriers
    doc.add_heading("Demographic information", level=2)
    info_df = pdf.loc[pdf.cases==True].groupby(
        ["sex_at_birth", "ancestry_pred_other"]
        ).agg({"person_id": "count"}).reset_index().rename(
            columns={"person_id": "n_individuals"}
    )
    doc.add_paragraph()  # spacing
    uth._add_table_to_doc(
        doc, info_df, uth._add_table_title(
            doc, f"Demographic information of {gene} homozygous pLoF carriers in AoU"), 
            info_df.columns.tolist(), list(map(Inches, [1, 1, 1]))
    )
    doc.add_paragraph()  # spacing
    # TODO: add clinical records, add labs and measurements
    # Survey table for homozygous carriers
    doc.add_heading("Survey responses", level=2)
    survey_df = pd.read_parquet(f"{BUCKET}/data/rwe_info/processed/selected_surveys.parquet")
    sdf = survey_df.loc[survey_df.person_id.isin(samples)].groupby("question").agg(
        {"answer": "value_counts"}
        ).rename(columns={"answer": "n_individuals"}).reset_index()
    doc.add_paragraph()  # spacing
    uth._add_table_to_doc(
        doc, sdf, uth._add_table_title(
            doc, f"Survey responses of {gene} homozygous pLoF carriers in AoU"), 
            sdf.columns.tolist(), list(map(Inches, [3, 2, 1]))
    )
    doc.add_paragraph()  # spacing
    doc.add_page_break()
    return doc

def get_person_information(sample_id, pdf):
    assert len(pdf.loc[pdf.person_id==sample_id])==1
    description_columns = ["sex_at_birth", "age_at_cdr", "ancestry_pred_other", "race", "ethnicity"]
    info_columns = ["has_ehr_data"]
    info_dict = pdf.loc[pdf.person_id==sample_id, description_columns + info_columns].iloc[0, :].to_dict()
    return info_dict

def get_variant_information(sample_id, vdf):
    variant_info_column = ["locus", "alleles", "gene", "consequence"]
    svdf = vdf.loc[vdf.homo_samples==sample_id].drop_duplicates(subset=variant_info_column).copy()
    svdf = svdf.loc[:, variant_info_column]
    info_dict = svdf.reset_index(drop=True).to_dict()
    return len(svdf), info_dict

def get_clinical_information(sample_id, phe_df, phe_info_df, ntop=10):
    phe_df = phe_df.loc[phe_df.person_id==sample_id].sort_values("count", ascending=False)
    nphenos = len(phe_df)
    phe_df = phe_df.head(ntop)
    phe_df = phe_df.merge(phe_info_df, left_on="phecode", right_on="PHECODE")
    return nphenos, phe_df.PHENOTYPE.unique()

def clean_numerical_measurements(df):
    df = df.loc[(df.unit_concept_id!=0)].copy()
    return df

def get_numerical_measurements(sample_id, nm_df, trait_range_dict):
    snm_df = nm_df.loc[
        (nm_df.person_id==sample_id)&(nm_df.measurement.isin(trait_range_dict.keys())),
        ["measurement", "unit", "median_value"]
    ].copy()
    snm_df["normal_range"] = snm_df.measurement.map(trait_range_dict)
    snm_df.columns = [c.capitalize().replace("_", " ") for c in snm_df.columns]
    return snm_df.sort_values("Measurement").rename(columns={"Median value": "Observed value"})

def get_survey_answers(sample_id, survey_df):
    sdf = survey_df.loc[survey_df.person_id==sample_id, ["question", "answer_category"]]
    sinfo_dict = dict(zip(sdf.question, sdf.answer_category))
    return sinfo_dict

def get_case_study_info_helper(doc, sample_id, pdf, vdf, phe_df, phe_info_df, nm_df, trait_range_dict, survey_df):
    ###--- Variant information and demographics ---###
    nvars, vinfo = get_variant_information(sample_id, vdf)
    case_study = doc.add_paragraph(
        f"An individual was identified in the AoU cohort who carries biallelic variant(s) "
        f"in {vinfo['gene'][0]} at loci {','.join(vinfo['locus'][i] for i in range(nvars))} with "
        f"{', '.join(vinfo['consequence'][i] for i in range(nvars))} consequence. "
    )
    person_info = get_person_information(sample_id, pdf)
    case_study.add_run(
        f"The individual is a {person_info['sex_at_birth']} of age {person_info['age_at_cdr']}. "
        f"Their genetic ancestry is predicted as {person_info['ancestry_pred_other']}. "
        f"They identify themselves of {person_info['race']} race and {person_info['ethnicity']} ethnicity. "
    )
    ###--- Clinical records ---###
    if person_info["has_ehr_data"]:
        case_study.add_run(f"EHR data of that individual is available. ")
        nphenos, phenos = get_clinical_information(sample_id, phe_df, phe_info_df)
        case_study.add_run(f"The individual has {nphenos} phenotypes in their clinical record. ")
        if nphenos>0:
            phenos = [f"{i+1}. {str(p).replace(',', ';')}" for i,p in enumerate(phenos)]
            case_study.add_run(f"The top phenotypes based on evidence include {', '.join(phenos)}. ")
    else:
        case_study.add_run(f"EHR data of that individual is not available.")
    ###--- Labs and measurements ---###
    snm_df = get_numerical_measurements(sample_id, nm_df, trait_range_dict)
    if len(snm_df)>0:
        uth._add_table_to_doc(
            doc, snm_df, uth._add_table_title(doc, f"Labs and measurement readings of the individual in AoU"),
                        ['Measurement', 'Unit', 'Observed value', 'Normal range'], list(map(Inches, [2.5, 1.25, 1, 1.25]))
                        )
    ###--- Surveys ---###
    sinfo_dict = get_survey_answers(sample_id, survey_df)
    if sinfo_dict:
        qa = [f"{i+1}. {q}: {a}" for i, (q,a) in enumerate(sinfo_dict.items())]
        case_study.add_run(f"The answers to the health related survey questions were: {', '.join(qa)}. ")
    return doc

def generate_aou_homozygous_lof_carriers_case_study(doc, chrm, gene, phe_info_file=""):
    from rwe.parsers.aou.config import BUCKET, NORMAL_TRAIT_RANGES
    # variant information and demographics
    pdf, vdf = get_individual_info(BUCKET, chrm, gene, zygosity="homo")
    samples = sorted(set(vdf.homo_samples.dropna().values))
    # clinical records
    phecode_df = pd.read_csv(f"{BUCKET}/data/phecodes/aou_phecodev1.2_counts.csv", dtype={"phecode": str})
    if not phe_info_file:
        assets_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'assets')
        phe_info_file = os.path.join(assets_dir, 'ICD_to_Phecode_mapping.csv')
    phe_info_df = pd.read_csv(phe_info_file, dtype=str)
    # labs and measurements
    numerical_measurements_df = pd.read_parquet(f"{BUCKET}/data/rwe_info/processed/selected_numerical_measurements.parquet")
    numerical_measurements_df = clean_numerical_measurements(numerical_measurements_df)
    # surveys
    survey_df = pd.read_parquet(f"{BUCKET}/data/rwe_info/processed/selected_surveys.parquet")
    survey_df = clean_aou_surveys(survey_df, zygosity="homo")
    doc.add_paragraph()  # spacing
    doc.add_heading("Case studies of homozygous LoF carriers in All of Us cohort", level=2)
    for i,sample in enumerate(samples):
        doc.add_paragraph()
        cs = doc.add_paragraph(f"Case Study {i+1}")
        cs.runs[0].bold = True
        doc = get_case_study_info_helper(doc, sample, pdf, vdf, phecode_df, phe_info_df, numerical_measurements_df, NORMAL_TRAIT_RANGES, survey_df)
    return doc

def generate_aou_homozygous_lof_carriers_report(doc: Document, chrm: str, gene: str) -> Document:
    from rwe.parsers.aou.config import BUCKET
    variant_df = pd.read_csv(f"{BUCKET}/data/rwe_info/genes/chr{chrm}/{gene}/lof_homo.csv.gz")
    if variant_df.empty:
        doc.add_paragraph(f"No AoU individuals with homozygous loss of function variants found for {gene}.")
        doc.add_page_break()  # spacing
    else:
        if variant_df["homo_samples"].dropna().nunique()>20:
            doc = generate_aou_homozygous_lof_carriers_report_summary(doc, chrm, gene)
        else:
            doc = generate_aou_homozygous_lof_carriers_case_study(doc, chrm, gene)
    doc.add_page_break()  # spacing
    return doc

@lru_cache(maxsize=1)
def _load_aou_ptv_counts() -> pd.DataFrame:
    resource = files("rwe") / "assets" / _AOU_PTV_COUNTS_RESOURCE
    with as_file(resource) as path:
        df = pd.read_csv(
            path,
            usecols=["gene", "nhetz_samples", "nhomz_samples", "chrm"],
        )
    df["gene"] = df["gene"].astype(str).str.upper()
    return df


def get_aou_carrier_info(gene: str) -> dict:
    normalized_gene = gene.strip().upper()
    df = _load_aou_ptv_counts()
    hits = df.loc[df["gene"] == normalized_gene]

    if hits.empty:
        nhetz = None
        nhomz = None
        chrm = None
    else:
        row = hits.iloc[0]
        nhetz = None if pd.isna(row["nhetz_samples"]) else int(row["nhetz_samples"])
        nhomz = None if pd.isna(row["nhomz_samples"]) else int(row["nhomz_samples"])
        chrm = None if pd.isna(row["chrm"]) else str(row["chrm"])

    return {
        "gene": normalized_gene,
        "chrm": chrm,
        "nhetz": nhetz,
        "nhomz": nhomz,
    }

if __name__ == "__main__":
    gene = "ANGPTL3"
    info = get_aou_carrier_info(gene)
    print(info) 



import os
from datetime import date
from importlib.metadata import version, PackageNotFoundError
from docx import Document
from docx.shared import Inches, Pt
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.oxml import OxmlElement
from docx.oxml.ns import qn
import pandas as pd


import rwe.utils.helpers as uth
import rwe.clients.aou as aou
import rwe.clients.azn as azn
import rwe.clients.genebass as gbs


def generate_title_and_contents(gene: str, logo_path: str | None = None, report_date: str | None = None) -> Document:
    """
    Generate the title and table of contents pages for the RWE report.

    Parameters
    ----------
    gene : str
        Gene symbol for the report (e.g., "PCSK9").
    logo_path : str | None
        Path to logo image file to include on title page. If None or file not found,
        falls back to default logo in assets folder. If no logo is available, no logo is added.
    report_date : str | None
        Date string to include on title page. If None, defaults to today's date in "Month DD, YYYY" format.

    Returns
    -------
    Document
        A python-docx Document object with the title and TOC pages added.
    """
    if not report_date:
        # default to today's date (you can override)
        report_date = date.today().strftime("%B %d, %Y")

    doc = Document()
    uth._set_default_styles(doc)
    uth._init_report_counters(doc)

    # --- Title page ---
    # Determine which logo to use
    final_logo_path = None
    if logo_path and os.path.exists(logo_path):
        final_logo_path = logo_path
    else:
        # Fallback to default logo in assets folder
        assets_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'assets')
        default_logo = os.path.join(assets_dir, 'Arrowhead_Pharmaceuticals_logo.png')
        if os.path.exists(default_logo):
            final_logo_path = default_logo

    if final_logo_path:
        p = doc.add_paragraph()
        p.alignment = WD_ALIGN_PARAGRAPH.CENTER
        # adjust width as needed
        p.add_run().add_picture(final_logo_path, width=Inches(2.0))

    title = doc.add_paragraph()
    title.alignment = WD_ALIGN_PARAGRAPH.CENTER
    r = title.add_run(f"Real World Evidence of {gene} gene pLoF carriers")
    r.bold = True
    r.font.size = Pt(24)

    doc.add_paragraph()  # spacer

    authors = doc.add_paragraph()
    authors.alignment = WD_ALIGN_PARAGRAPH.CENTER
    authors.add_run("Deepro Banerjee, Shicheng Guo").font.size = Pt(14)

    dt = doc.add_paragraph()
    dt.alignment = WD_ALIGN_PARAGRAPH.CENTER
    dt.add_run(report_date).font.size = Pt(14)

    dept = doc.add_paragraph()
    dept.alignment = WD_ALIGN_PARAGRAPH.CENTER
    dept.add_run("Translational Genetics and Data Science").font.size = Pt(14)

    org = doc.add_paragraph()
    org.alignment = WD_ALIGN_PARAGRAPH.CENTER
    org.add_run("Arrowhead Pharmaceuticals").font.size = Pt(14)

    doc.add_page_break()

    # --- Contents page ---
    p = doc.add_paragraph()
    r = p.add_run("Table of Contents")
    r.bold = True
    r.font.size = Pt(14)

    # TOC field goes in its own empty paragraph
    toc_p = doc.add_paragraph()
    uth._add_toc_field(toc_p)


    note = doc.add_paragraph(
        "Note: In Microsoft Word, right-click the table of contents and choose 'Update Field' "
        "to populate page numbers."
    )
    note.runs[0].italic = True

    doc.add_page_break()
    return doc


def generate_variant_information_and_demographics(doc: Document, chrm: str, gene: str, zygosity: str, allofus: bool = True) -> Document:
    doc.add_heading("Variant information and demographics", level=1)
    # AoU
    if allofus:
        doc = aou.generate_aou_variant_info_demographics_report(doc, chrm, gene, zygosity)   
    return doc

def generate_clinical_records(doc: Document, chrm:str, gene: str, zygosity: str,  output_dir: str, allofus: bool = True, genebass: bool = False, astrazeneca: bool = True) -> Document:
    doc.add_heading("Clinical records", level=1)
    # AoU
    if allofus:
        doc = aou.generate_aou_clinical_report(doc, chrm, gene, zygosity, output_dir)
    # GeneBass
    if genebass:
        doc = gbs.generate_genebass_clinical_report(doc, gene, output_dir)
    # AZN
    if astrazeneca:
        doc = azn.generate_azn_clinical_report(doc, gene, output_dir)
    return doc

def generate_labs_and_measurements(doc: Document, chrm: str, gene: str, zygosity: str, output_dir: str, allofus: bool = True, genebass: bool = False, astrazeneca: bool = True) -> Document:
    doc.add_heading("Labs and measurements", level=1)
    # AoU
    if allofus:
        doc = aou.generate_aou_labs_measurements_report(doc, chrm, gene, zygosity)
        doc.add_paragraph("")
    # GeneBass
    if genebass:
        doc = gbs.generate_genebass_labs_measurements_report(doc, gene, output_dir)
        doc.add_paragraph("")
    # AZN
    if astrazeneca:
        doc = azn.generate_azn_labs_measurements_report(doc, gene, output_dir)
        doc.add_paragraph("")
    return doc

def generate_survey_information(doc: Document, chrm: str, gene: str, zygosity: str, output_dir: str, allofus: bool = True) -> Document:
    doc.add_heading("Survey information", level=1)
    # AoU
    if allofus:
        doc = aou.generate_aou_survey_report(doc, chrm, gene, zygosity)
        doc.add_paragraph("")
    return doc

def generate_homozygous_lof_carriers(doc: Document, chrm: str, gene: str, output_dir: str, allofus: bool = True) -> Document:
    doc.add_heading("Homozygous loss of function carriers", level=1)
    # AoU
    if allofus:
        doc = aou.generate_aou_homozygous_lof_carriers_report(doc, chrm, gene)
        doc.add_paragraph("")
    return doc

def generate_indication_specific_report(doc: Document, chrm: str, gene: str, output_dir: str, indications: list, keywords: list, allofus: bool = True, genebass: bool = False, astrazeneca: bool = True) -> Document:
    doc.add_heading(f"Indication-specific report for {gene}", level=1)
    doc.add_paragraph("Indication(s) searched: " + ", ".join(indications))
    # AoU
    if allofus:
        doc = aou.generate_aou_indication_specific_report(doc, chrm, gene, output_dir, keywords)
    # GeneBass
    if genebass:
        doc = gbs.generate_genebass_indication_specific_report(doc, gene, output_dir, keywords)
    # AZN     
    if astrazeneca:
        doc = azn.generate_azn_indication_specific_report(doc, gene, output_dir, keywords)
    doc.add_paragraph("")
    return doc

def generate_plasma_proteomics_report(doc: Document, chrm: str, gene: str, output_dir: str, astrazeneca: bool = True) -> Document:
    doc.add_heading("Plasma proteomics", level=1)
    # AZN
    if astrazeneca:
        doc = azn.generate_azn_plasma_protein_biomarkers_report(doc, gene, output_dir)
    doc.add_paragraph("")
    return doc

def generate_end_of_report(doc: Document) -> Document:
    try:
        pkg_version = version("rwe")
    except PackageNotFoundError:
        pkg_version = "unknown (local source)"
    doc.add_heading("End of report", level=1)
    doc.add_paragraph()
    note = doc.add_paragraph(
        f"This report was generated by rwe package v{pkg_version} using data available at the time of analysis."
    )
    note.runs[0].italic = True
    return doc



import os
from datetime import date
from docx import Document
from docx.shared import Inches, Pt
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.oxml import OxmlElement
from docx.oxml.ns import qn
from google.cloud import storage

_FIGURE_SAVE_DIR = ""
_FIGURE_SAVE_PREFIX = ""

# ---------- DOCX helpers ----------

def _set_figure_save_context(output_dir: str, prefix: str = "") -> None:
    global _FIGURE_SAVE_DIR, _FIGURE_SAVE_PREFIX
    _FIGURE_SAVE_DIR = output_dir
    _FIGURE_SAVE_PREFIX = prefix


def _clear_figure_save_context() -> None:
    global _FIGURE_SAVE_DIR, _FIGURE_SAVE_PREFIX
    _FIGURE_SAVE_DIR = ""
    _FIGURE_SAVE_PREFIX = ""

def _init_report_counters(doc: Document) -> Document:
    """Initialize figure and table counters on the document object."""
    doc._report_counters = {"figure": 0, "table": 0}
    return doc


def _get_next_figure_num(doc: Document) -> int:
    """Get the next figure number and increment the counter."""
    if not hasattr(doc, '_report_counters'):
        doc._report_counters = {"figure": 0, "table": 0}
    doc._report_counters["figure"] += 1
    return doc._report_counters["figure"]


def _get_next_table_num(doc: Document) -> int:
    """Get the next table number and increment the counter."""
    if not hasattr(doc, '_report_counters'):
        doc._report_counters = {"figure": 0, "table": 0}
    doc._report_counters["table"] += 1
    return doc._report_counters["table"]


def _set_default_styles(doc: Document):
    # Normal
    style = doc.styles["Normal"]
    style.font.name = "Calibri"
    style.font.size = Pt(11)

    # Headings
    for hname, size in [("Heading 1", 14), ("Heading 2", 12)]:
        s = doc.styles[hname]
        s.font.name = "Calibri"
        s.font.size = Pt(size)
        s.font.bold = True


def _add_toc_field(paragraph):
    """
    Insert a Word TOC field. Word will render it after you:
      References -> Update Table
    """
    run = paragraph.add_run()

    fld_begin = OxmlElement("w:fldChar")
    fld_begin.set(qn("w:fldCharType"), "begin")

    instr = OxmlElement("w:instrText")
    instr.set(qn("xml:space"), "preserve")
    # TOC switches:
    #   \\o "1-3" include headings 1-3
    #   \\h hyperlinks
    #   \\z hide page numbers in web layout
    #   \\u use outline levels
    instr.text = r'TOC \o "1-3" \h \z \u'

    fld_sep = OxmlElement("w:fldChar")
    fld_sep.set(qn("w:fldCharType"), "separate")

    fld_end = OxmlElement("w:fldChar")
    fld_end.set(qn("w:fldCharType"), "end")

    run._r.append(fld_begin)
    run._r.append(instr)
    run._r.append(fld_sep)
    run._r.append(fld_end)
    return


def _add_caption(doc: Document, text: str):
    """Add a caption paragraph (for backward compatibility)."""
    p = doc.add_paragraph(text)
    p.style = doc.styles["Normal"]
    p.runs[0].italic = True
    p.alignment = WD_ALIGN_PARAGRAPH.CENTER
    return p


def _add_figure_caption(doc: Document, description: str):
    """Add an auto-numbered figure caption.

    Args:
        doc: Document object with _report_counters attached
        description: Caption text without the "Figure X." prefix
    """
    fig_num = _get_next_figure_num(doc)
    text = f"Figure {fig_num}. {description}"
    return _add_caption(doc, text)


def _add_table_title(doc: Document, description: str):
    """Get an auto-numbered table title string.

    Args:
        doc: Document object with _report_counters attached
        description: Title text without the "Table X." prefix

    Returns:
        str: The full table title with number prefix
    """
    table_num = _get_next_table_num(doc)
    return f"Table {table_num}. {description}"

def _save_fig_to_tmp(fig, basename="figure", dpi=300):
    # Keep function name for backward compatibility, but always persist figures.
    fig_dir = _FIGURE_SAVE_DIR
    if not fig_dir:
        fallback_gene_dir = _FIGURE_SAVE_PREFIX or "gene"
        fig_dir = os.path.join(".", "data", "rwe_reports", fallback_gene_dir)
    os.makedirs(fig_dir, exist_ok=True)
    fig_prefix = _FIGURE_SAVE_PREFIX
    fig_name = f"{fig_prefix}_{basename}.png" if fig_prefix else f"{basename}.png"
    path = os.path.join(fig_dir, fig_name)
    fig.savefig(path, dpi=dpi, bbox_inches="tight")
    return path

def _set_table_autofit(table, autofit: bool = False):
    # python-docx doesn't expose this cleanly; set tblLayout to fixed
    tbl_pr = table._tbl.tblPr
    tbl_layout = tbl_pr.find(qn('w:tblLayout'))
    if tbl_layout is None:
        tbl_layout = OxmlElement('w:tblLayout')
        tbl_pr.append(tbl_layout)
    tbl_layout.set(qn('w:type'), 'autofit' if autofit else 'fixed')

def _add_table_to_doc(doc, df, title, columns, col_widths=None, table_width_inches=None):
    """
    col_widths: list same length as columns, items are Inches/Cm, e.g. [Inches(1.2), Inches(3.0), ...]
    table_width_inches: optional total width; if provided and col_widths is None, widths are evenly split
    """
    # Add title
    p = doc.add_paragraph()
    run = p.add_run(title)
    run.bold = True

    table = doc.add_table(rows=1, cols=len(columns))
    table.style = 'Light Grid Accent 1'
    _set_table_autofit(table, autofit=False)

    # decide widths
    if col_widths is None and table_width_inches is not None:
        w = Inches(table_width_inches / len(columns))
        col_widths = [w] * len(columns)

    # Header
    header_cells = table.rows[0].cells
    for i, col in enumerate(columns):
        header_cells[i].text = col
        header_cells[i].paragraphs[0].runs[0].bold = True
        if col_widths is not None:
            header_cells[i].width = col_widths[i]

    # Rows
    for _, row in df.iterrows():
        row_cells = table.add_row().cells
        for i, col in enumerate(columns):
            value = row[col]
            if isinstance(value, float):
                if col in ('P-Value (Burden)', 'P-value', 'P value', 'p_value'):
                    row_cells[i].text = f"{value:.2e}"
                else:
                    row_cells[i].text = f"{value:.2f}"
            else:
                row_cells[i].text = str(value)

            if col_widths is not None:
                row_cells[i].width = col_widths[i]

    doc.add_paragraph()
    return doc

# AoU helpers
def _relpath(gs_uri: str, bucket: str) -> str:
    # gs://bucket/prefix/file -> prefix/file
    return gs_uri.replace(f"{bucket}/", "")

def _gcs_size(gs_uri: str, bucket: str) -> int:
    storage_client = storage.Client()
    bucket_obj = storage_client.bucket(bucket.lstrip("gs://"))
    rel = _relpath(gs_uri, bucket)
    blob = bucket_obj.blob(rel)
    if not blob.exists():
        return 0
    blob.reload()
    return blob.size or 0



import os
from datetime import date, datetime
from importlib.metadata import PackageNotFoundError, version
from importlib.resources import as_file, files

import rwe.clients.aou as aou
import rwe.clients.genebass as gbs
import rwe.clients.gnomad as gnomad


_DEFAULT_TEMPLATE_RESOURCE = "assets/RWE_template.pptx"

_DEFAULT_SLIDE4_ANCHORS = {
    "aou": (306000, 1110629, 5315528, 2686727),
    "ukb": (6276035, 1114132, 5419390, 2732276),
    "survey": (599025, 4035903, 4688785, 2360422),
}


def _send_shape_to_back(slide, shape) -> None:
    sp_tree = slide.shapes._spTree
    sp = shape._element
    sp_tree.remove(sp)
    sp_tree.insert(2, sp)


def _replace_title_slide_text(slide, gene: str, indications: list[str], report_date: str) -> None:
    from pptx.util import Pt
    cleaned_indications = [indication.replace("_", " ").strip() for indication in indications if indication.strip()]
    indication_text = ", ".join(cleaned_indications) if cleaned_indications else "general"
    for shape in slide.shapes:
        if not getattr(shape, "has_text_frame", False):
            continue
        text = shape.text.strip()
        if "Real world evidence for" in text:
            text_frame = shape.text_frame
            title_font_size = Pt(60)
            text_frame.clear()
            p = text_frame.paragraphs[0]
            lead_run = p.add_run()
            lead_run.text = "Real world evidence for "
            lead_run.font.size = title_font_size
            gene_run = p.add_run()
            gene_run.text = gene
            gene_run.font.italic = True
            gene_run.font.size = title_font_size
            suffix_run = p.add_run()
            suffix_run.text = f" - {indication_text}"
            suffix_run.font.size = title_font_size
        elif "Deepro Banerjee" in text and "Arrowhead Pharmaceuticals" in text:
            shape.text = (
                "Deepro Banerjee, Shicheng Guo\n"
                f"{report_date}\n"
                "Translational Genetics and Data Science\n"
                "Arrowhead Pharmaceuticals"
            )
    return

def _extract_slide4_anchors(slide, mso_shape_type):
    picture_shapes = [s for s in slide.shapes if s.shape_type == mso_shape_type.PICTURE]
    picture_shapes = sorted(picture_shapes, key=lambda s: (s.top, s.left))
    if len(picture_shapes) >= 3:
        return {
            "aou": (picture_shapes[0].left, picture_shapes[0].top, picture_shapes[0].width, picture_shapes[0].height),
            "ukb": (picture_shapes[1].left, picture_shapes[1].top, picture_shapes[1].width, picture_shapes[1].height),
            "survey": (picture_shapes[2].left, picture_shapes[2].top, picture_shapes[2].width, picture_shapes[2].height),
        }
    return _DEFAULT_SLIDE4_ANCHORS.copy()


def _format_count_with_breakdown(values: list[tuple[str, int | None]]) -> str:
    present = [(label, value) for label, value in values if value is not None]
    if not present:
        return "NA"
    total = sum(value for _, value in present)
    if total == 0:
        return "0"
    breakdown = " + ".join(str(value) for _, value in present)
    return f"{total} ({breakdown})"


def _build_safety_bullets(gene: str, output_dir: str, include_aou: bool, include_genebass: bool) -> list[str]:
    hgnc_path = os.path.join(output_dir, "gene_with_protein_product.txt")
    constraint_path = os.path.join(output_dir, "gnomad.v4.1.constraint_metrics.tsv")

    gnomad_info = gnomad.get_gnomad_constraint_info(
        gene,
        constraint_path=constraint_path,
        hgnc_path=hgnc_path,
    )
    approved_gene = gnomad_info["gene"]
    genebass_info = gbs.get_genebass_carrier_info(approved_gene) if include_genebass else {"nhetz": None, "nhomz": None}
    aou_info = aou.get_aou_carrier_info(gene) if include_aou else {"nhetz": None, "nhomz": None}

    heterozygous = [
        ("gnomAD", gnomad_info.get("nhetz")),
        ("UK Biobank", genebass_info.get("nhetz")),
        ("All of Us", aou_info.get("nhetz")),
    ]
    homozygous = [
        ("gnomAD", gnomad_info.get("nhomz")),
        ("UK Biobank", genebass_info.get("nhomz")),
        ("All of Us", aou_info.get("nhomz")),
    ]
    included_labels = [label for label, value in heterozygous if value is not None] + [
        label for label, value in homozygous if value is not None and label not in [l for l, _ in heterozygous if _ is not None]
    ]
    if not included_labels:
        cohort_text = "gnomAD, UK Biobank and All of Us"
    elif len(included_labels) == 1:
        cohort_text = included_labels[0]
    elif len(included_labels) == 2:
        cohort_text = f"{included_labels[0]} and {included_labels[1]}"
    else:
        cohort_text = ", ".join(included_labels[:-1]) + f" and {included_labels[-1]}"

    loeuf = gnomad_info.get("loeuf")
    loeuf_text = "NA" if loeuf is None else f"{loeuf:.3f}"

    return [
        "No consistent cross-biobank toxicity signal in clinical records of predicted loss of function carriers.",
        f"LOEUF score: {loeuf_text}.",
        (
            f"{cohort_text} cohorts have {_format_count_with_breakdown(heterozygous)} heterozygous "
            f"and {_format_count_with_breakdown(homozygous)} homozygous loss of function carriers."
        ),
        "Self-reported survey information of four health related questions in loss of function carriers is similar to non-carriers.",
    ]


def _update_slide4_safety_text(slide, bullets: list[str]) -> None:
    from pptx.util import Pt

    for shape in slide.shapes:
        if not getattr(shape, "has_text_frame", False):
            continue
        text = shape.text.strip()
        if "No consistent cross-biobank toxicity signal" not in text:
            continue
        text_frame = shape.text_frame
        paragraphs = list(text_frame.paragraphs)
        while len(paragraphs) < len(bullets):
            paragraphs.append(text_frame.add_paragraph())
        for idx, bullet in enumerate(bullets):
            paragraph = paragraphs[idx]
            paragraph.text = bullet
            for run in paragraph.runs:
                run.font.size = Pt(14)
        for paragraph in paragraphs[len(bullets):]:
            paragraph.text = ""
        return


def _populate_slide4(
    slide,
    gene: str,
    output_dir: str,
    aou_plot: str | None,
    ukb_plot: str | None,
    survey_plot: str | None,
    include_aou: bool,
    include_genebass: bool,
    mso_shape_type,
) -> None:
    anchors = _extract_slide4_anchors(slide, mso_shape_type)

    for shape in list(slide.shapes):
        if shape.shape_type == mso_shape_type.PICTURE:
            shape.element.getparent().remove(shape.element)
    if aou_plot and os.path.exists(aou_plot):
        left, top, width, height = anchors["aou"]
        aou_picture = slide.shapes.add_picture(aou_plot, left, top, width=width, height=height)
        _send_shape_to_back(slide, aou_picture)
    if ukb_plot and os.path.exists(ukb_plot):
        left, top, width, height = anchors["ukb"]
        ukb_picture = slide.shapes.add_picture(ukb_plot, left, top, width=width, height=height)
        _send_shape_to_back(slide, ukb_picture)
    if survey_plot and os.path.exists(survey_plot):
        left, top, width, height = anchors["survey"]
        survey_picture = slide.shapes.add_picture(survey_plot, left, top, width=width, height=height)
        _send_shape_to_back(slide, survey_picture)
    bullets = _build_safety_bullets(gene, output_dir, include_aou, include_genebass)
    _update_slide4_safety_text(slide, bullets)
    return

def generate_rwe_presentation(
    gene: str,
    indications: str | list[str] | None = None,
    output_dir: str = "",
    report_date: str = "",
    allofus: bool = True,
    astrazeneca: bool = True,
    genebass: bool = True,
) -> str:
    """
    Generate a 6-slide RWE PPTX from the template.

    Slide behavior:
    1. Title slide: update gene, indications, and date.
    2. Keep as template.
    3. Keep as template.
    4. Clinical safety slide with AoU + UKB (AZN preferred, GeneBass fallback) Manhattan plots and AoU survey plot.
       These plots are expected to already exist in output_dir from the DOCX generation step.
    5. Keep as template.
    6. Keep as template.
    """
    try:
        from pptx import Presentation
        from pptx.enum.shapes import MSO_SHAPE_TYPE
    except ImportError as exc:
        raise ImportError(
            "python-pptx is required to generate PowerPoint files. Install with: pip install python-pptx"
        ) from exc

    if isinstance(indications, str):
        indications_list = [x for x in indications.split(",") if x]
    else:
        indications_list = indications or []
    if not report_date:
        report_date = date.today().strftime("%B %d, %Y")

    if not output_dir:
        output_dir = "./rwe_reports"
    os.makedirs(output_dir, exist_ok=True)

    date_tag = datetime.today().strftime("%m%d%y")
    indication_part = "-".join(indications_list) if indications_list else "general"
    try:
        pkg_version = version("rwe")
    except PackageNotFoundError:
        pkg_version = "unknown"
    out_pptx_path = os.path.join(
        output_dir,
        f"RWE_{gene}_{indication_part}_report_{date_tag}_v{pkg_version}.pptx",
    )

    aou_plot = None
    if allofus:
        candidate = os.path.join(output_dir, f"{gene}_aou_phewas.png")
        if os.path.exists(candidate):
            aou_plot = candidate

    ukb_plot = None
    if astrazeneca:
        candidate = os.path.join(output_dir, f"{gene}_azn_phewas.png")
        if os.path.exists(candidate):
            ukb_plot = candidate
    if ukb_plot is None and genebass:
        candidate = os.path.join(output_dir, f"{gene}_genebass_phewas.png")
        if os.path.exists(candidate):
            ukb_plot = candidate

    survey_plot = None
    if allofus:
        candidate = os.path.join(output_dir, f"{gene}_aou_surveys.png")
        if os.path.exists(candidate):
            survey_plot = candidate
        else:
            candidate = os.path.join(output_dir, f"{gene}_aou_survey.png")
            if os.path.exists(candidate):
                survey_plot = candidate

    template_resource = files("rwe").joinpath(_DEFAULT_TEMPLATE_RESOURCE)
    with as_file(template_resource) as template_fs_path:
        prs = Presentation(str(template_fs_path))

    if len(prs.slides) < 6:
        raise ValueError("Template must contain at least 6 slides.")

    _replace_title_slide_text(prs.slides[0], gene, indications_list, report_date)
    _populate_slide4(
        prs.slides[3],
        gene,
        output_dir,
        aou_plot,
        ukb_plot,
        survey_plot,
        allofus,
        genebass,
        MSO_SHAPE_TYPE,
    )

    prs.save(out_pptx_path)
    return out_pptx_path



import os
from datetime import datetime

import matplotlib
matplotlib.use("Agg")  # important for batch/report generation (no display)

import rwe.utils.helpers as uth
import rwe.utils.report as utr
import rwe.utils.presentation as upp
from rwe.utils.keywords import disease_to_keywords


# ---------- Main user-facing function ----------

def generate_rwe_report(
    gene: str,
    chrm: str,
    out_docx_path: str = "",
    logo_path: str = "",
    report_date: str = "",
    indications: str = "",
    allofus: bool = True,
    genebass: bool = True,
    astrazeneca: bool = True,
    generate_pptx: bool = True,
):
    """
    Creates a DOCX report and optionally a PPTX report with:
      - Title page
      - Contents (TOC field)
      - Five section headers
      - Variant information + Demographics figures inserted into section 1

    Parameters
    ----------
    gene : str
    chrm : str
    out_docx_path : str
    logo_path : optional str (path to Arrowhead logo PNG/JPG)
    report_date : optional str (e.g. "January 26th, 2026")
    indications : optional str (e.g. "obesity,diabetes")
    generate_pptx : bool
        If True, also generate a PowerPoint summary using the template.
    """
    if not out_docx_path:
        date_tag = datetime.today().strftime("%m%d%y")
        out_docx_path = f"./data/rwe_reports/{gene}/RWE_{gene}_report_{date_tag}.docx"
    proj_dir = os.path.dirname(out_docx_path) or f"./data/rwe_reports/{gene}"
    os.makedirs(proj_dir, exist_ok=True)
    uth._set_figure_save_context(proj_dir, gene)

    try:
        # --- Title and Contents --- #
        doc = utr.generate_title_and_contents(gene, logo_path=logo_path, report_date=report_date)

        # --- Variant information and demographics --- #
        doc = utr.generate_variant_information_and_demographics(doc, chrm, gene, "hetz", allofus=allofus)

        # --- Clinical records --- #
        doc = utr.generate_clinical_records(doc, chrm, gene, "hetz", proj_dir, allofus=allofus, genebass=genebass, astrazeneca=astrazeneca)
        
        # --- Labs and measurements --- #
        doc = utr.generate_labs_and_measurements(doc, chrm, gene, "hetz", proj_dir, allofus=allofus, genebass=genebass, astrazeneca=astrazeneca)

        # --- Survey information --- #
        doc = utr.generate_survey_information(doc, chrm, gene, "hetz", proj_dir, allofus=allofus)

        # --- Homozygous loss of function carriers --- #
        doc = utr.generate_homozygous_lof_carriers(doc, chrm, gene, proj_dir, allofus=allofus)

        # --- Plasma proteomics --- #
        doc = utr.generate_plasma_proteomics_report(doc, chrm, gene, proj_dir, astrazeneca=astrazeneca)

        # --- Indication-specific report --- #
        indications_list = []
        if indications:
            indications_list = [i.strip() for i in indications.split(",")]
            keywords = sum([disease_to_keywords[i] for i in indications_list if i in disease_to_keywords], [])
            if keywords:
                doc = utr.generate_indication_specific_report(doc, chrm, gene, proj_dir, indications_list, keywords, allofus=allofus, genebass=genebass, astrazeneca=astrazeneca)
            else:
                print(f"Warning: no keywords found for indications {indications_list}; skipping indication-specific report")

        # --- End of report --- #
        doc = utr.generate_end_of_report(doc)

        # Save
        doc.save(out_docx_path)

        out_pptx_final = ""
        if generate_pptx:
            out_pptx_final = upp.generate_rwe_presentation(
                gene=gene,
                indications=indications_list if indications_list else indications,
                output_dir=proj_dir,
                report_date=report_date,
                allofus=allofus,
                astrazeneca=astrazeneca,
                genebass=genebass,
            )
        return {"docx_path": out_docx_path, "pptx_path": out_pptx_final}
    finally:
        uth._clear_figure_save_context()

if __name__ == "__main__":
    pass



import tempfile
import unittest
from pathlib import Path

from docx import Document

from rwe.clients.azn import (
    generate_azn_clinical_report,
    generate_azn_indication_specific_report,
    generate_azn_labs_measurements_report,
    generate_azn_plasma_protein_biomarkers_report,
)
from tests._fixtures import write_azn_binary_csv, write_azn_continuous_csv


class AstraZenecaReportSmokeTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.output_dir = Path(self.tmpdir.name)
        self.gene = "TESTGENE"
        write_azn_binary_csv(self.output_dir, self.gene)
        write_azn_continuous_csv(self.output_dir, self.gene)

    def tearDown(self) -> None:
        self.tmpdir.cleanup()

    def test_azn_report_smoke(self) -> None:
        doc = Document()
        doc = generate_azn_clinical_report(doc, self.gene, str(self.output_dir))
        doc = generate_azn_labs_measurements_report(doc, self.gene, str(self.output_dir))
        doc = generate_azn_plasma_protein_biomarkers_report(doc, self.gene, str(self.output_dir))
        doc = generate_azn_indication_specific_report(doc, self.gene, str(self.output_dir), ["diabetes", "obesity"])

        out_path = self.output_dir / f"{self.gene}_azn.docx"
        doc.save(out_path)

        self.assertTrue(out_path.exists())
        full_text = "\n".join(p.text for p in doc.paragraphs)
        self.assertIn("AstraZeneca", full_text)
        self.assertIn("Indication specific PheWAS results in UK Biobank cohort collected from AstraZeneca", full_text)


if __name__ == "__main__":
    unittest.main()



import pandas as pd
import numpy as np

# plotting imports 
import textwrap
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt

matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42
# matplotlib.rcParams['font.sans-serif'] = "Arial" # missing fonts:: https://alexanderlabwhoi.github.io/post/2021-03-missingfont/
# Then, "ALWAYS use sans-serif fonts"
matplotlib.rcParams['font.family'] = "sans-serif"
matplotlib.rcParams.update({'font.size': 6, 'axes.linewidth': 1, 'xtick.major.width': 1, 'xtick.major.size': 5, 'ytick.major.width': 1, 'ytick.major.size': 5})


def p_to_stars(p):
    if pd.isna(p): return ""
    if p < 1e-3: return "***"
    if p < 1e-2: return "**"
    if p < 5e-2: return "*"
    return ""

def fmt_p(p):
    if pd.isna(p): return ""
    return f"P={p:.2e}" if p < 0.001 else f"P={p:.3f}"

def plot_measurements(
    df_long,
    measurements,
    col="median_value",
    res_df=None,
    res_name_col="measurement",
    res_p_col="pvalue",
    ncols=3,
    figsize=(6, 4),
    palette=None
    ):
    d = df_long[df_long["measurement"].isin(measurements)].copy()
    d[col] = pd.to_numeric(d[col], errors="coerce")
    d = d.dropna(subset=[col])
    d["group"] = np.where(d["cases"], "Cases", "Controls")

    if palette is None:
        palette = {
            "Controls": "#2F6690",  # BLUE
            "Cases":    "#D1495B",  # RED
        }
    # star map
    p_map, star_map = {}, {}
    if res_df is not None:
        tmp = res_df.copy()
        tmp[res_p_col] = pd.to_numeric(tmp[res_p_col], errors="coerce")
        p_map = dict(zip(tmp[res_name_col], tmp[res_p_col]))
        star_map = {k: p_to_stars(v) for k, v in p_map.items()}
        
    # Layout: dynamic grid
    n = len(measurements)
    nrows = int(np.ceil(n / ncols))
    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)
    axes = np.array(axes).ravel()

    for i, m in enumerate(measurements):
        ax = axes[i]
        sub = d[d["measurement"] == m].copy()

        sns.violinplot(
            data=sub, x="group", y=col, order=["Controls", "Cases"],
            hue="group", hue_order=["Controls", "Cases"], legend=False,
            palette=palette, inner="quartile", cut=0, linewidth=1, ax=ax
            )

        ax.set_title(textwrap.fill(m, width=20,  max_lines=3))
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.spines[['top', 'right']].set_visible(False)
        
        stars = star_map.get(m, "")
        pval = p_map.get(m, np.nan)

        if stars or (not pd.isna(pval)):
            ymax = np.nanmax(sub[col].values)
            ymin = np.nanmin(sub[col].values)
            span = (ymax - ymin) if ymax > ymin else 1.0

            # put annotation above violins
            y_txt = ymax + 0.08 * span
            
            # x=0.5 is between the two categories (0 and 1)
            if stars:
                ax.text(0.5, y_txt + 0.03 * span, stars, ha="center", va="bottom")
            if not pd.isna(pval):
                ax.text(0.5, y_txt, fmt_p(pval), ha="center", va="bottom")

            # give headroom so text doesn't clip
            ax.set_ylim(ymin, y_txt + 0.12 * span)

    for j in range(n, len(axes)):
        axes[j].axis("off")

    plt.tight_layout()
    return fig, axes



import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

from docx import Document

from rwe.clients.genebass import (
    GeneBassClient,
    generate_genebass_clinical_report,
    generate_genebass_indication_specific_report,
    generate_genebass_labs_measurements_report,
)
from tests._fixtures import write_genebass_csv


class GeneBassReportSmokeTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.output_dir = Path(self.tmpdir.name)
        self.gene = "TESTGENE"
        write_genebass_csv(self.output_dir, self.gene)

    def tearDown(self) -> None:
        self.tmpdir.cleanup()

    def test_genebass_report_smoke(self) -> None:
        with patch("rwe.clients.genebass.get_gene_info", return_value=(self.gene, "ENSG_TEST", "1")):
            doc = Document()
            doc = generate_genebass_clinical_report(doc, self.gene, str(self.output_dir))
            doc = generate_genebass_labs_measurements_report(doc, self.gene, str(self.output_dir))
            doc = generate_genebass_indication_specific_report(doc, self.gene, str(self.output_dir), ["diabetes", "obesity"])

        out_path = self.output_dir / f"{self.gene}_genebass.docx"
        doc.save(out_path)

        self.assertTrue(out_path.exists())
        full_text = "\n".join(p.text for p in doc.paragraphs)
        self.assertIn("Phenome-wide Association Study (PheWAS) in UK Biobank cohort collected from GeneBass", full_text)
        self.assertIn("Indication specific PheWAS results in UK Biobank cohort collected from GeneBass", full_text)

    def test_genebass_carrier_info_smoke(self) -> None:
        info = GeneBassClient(gene=self.gene, phewas_dir="").get_carrier_info()
        self.assertEqual(info["gene"], self.gene)
        self.assertIn("nhetz", info)
        self.assertIn("nhomz", info)


if __name__ == "__main__":
    unittest.main()



import os

# get GC bucket and CDR info
BUCKET = os.getenv("WORKSPACE_BUCKET")
CDR = os.environ["WORKSPACE_CDR"]
GOOGLE_PROJECT = os.getenv("GOOGLE_PROJECT")


MEASUREMENT_GROUPS = {
    "vitals": [
        "Body mass index (BMI) [Ratio]", "Body height", "Body weight", 
        "Systolic blood pressure", "Diastolic blood pressure" , "Heart rate"
    ],
    "hematology": [
        "Hemoglobin [Mass/volume] in Blood",
        "Hematocrit [Volume Fraction] of Blood by Automated count",
        "Erythrocytes [#/volume] in Blood by Automated count",
        "Erythrocyte distribution width [Entitic volume]",
        "MCV [Entitic volume] by Automated count",
        "Platelets [#/volume] in Blood by Automated count",
    ],
    "renal": [
        "Creatinine [Mass/volume] in Serum or Plasma",
        "Urea nitrogen [Mass/volume] in Serum or Plasma",
        "Glomerular filtration rate/1.73 sq M.predicted [Volume Rate/Area] in Serum, Plasma or Blood by Creatinine-based formula (MDRD)",
        "Sodium [Moles/volume] in Serum or Plasma",
        "Potassium [Moles/volume] in Serum or Plasma",
        "Bicarbonate [Moles/volume] in Serum or Plasma",
    ],
    "liver": [
        "Albumin [Mass/volume] in Serum or Plasma",
        "Protein [Mass/volume] in Serum or Plasma",
        "Alanine aminotransferase [Enzymatic activity/volume] in Serum or Plasma",
        "Aspartate aminotransferase [Enzymatic activity/volume] in Serum or Plasma",
        "Alkaline phosphatase [Enzymatic activity/volume] in Serum or Plasma",
        "Bilirubin.total [Mass/volume] in Serum or Plasma",
    ],
    "metabolic": [
        "Glucose [Mass/volume] in Serum or Plasma",
        "Hemoglobin A1c/Hemoglobin.total in Blood",
        "Cholesterol [Mass/volume] in Serum or Plasma",
        "Cholesterol in LDL [Mass/volume] in Serum or Plasma by calculation",
        "Cholesterol in HDL [Mass/volume] in Serum or Plasma",
        "Triglyceride [Mass/volume] in Serum or Plasma",
    ],
    "miscellaneous": [
        "C reactive protein [Mass/volume] in Serum or Plasma",
        "Erythrocyte sedimentation rate by Westergren method",
        "Natriuretic peptide B [Mass/volume] in Serum or Plasma",
        "Natriuretic peptide.B prohormone N-Terminal [Mass/volume] in Serum or Plasma",
        "Troponin I.cardiac [Mass/volume] in Serum or Plasma",
        "Creatine kinase [Enzymatic activity/volume] in Serum or Plasma",
    ]
}

NORMAL_TRAIT_RANGES = {
    # Obesity / vital signs
    "Body mass index (BMI) [Ratio]": (18.5, 24.9),  # kg/m^2
    "Systolic blood pressure": (90, 120),           # mmHg
    "Diastolic blood pressure": (60, 80),           # mmHg
    "Heart rate": (60, 100),                        # bpm

    # Diabetes / glycemia
    "Glucose [Mass/volume] in Serum or Plasma": (70, 99),                 # fasting mg/dL
    "Hemoglobin A1c/Hemoglobin.total in Blood": (4.0, 5.6),               # %

    # Lipids / CVD risk
    "Cholesterol [Mass/volume] in Serum or Plasma": (0, 200),             # total mg/dL (desirable <200)
    "Cholesterol in LDL [Mass/volume] in Serum or Plasma by calculation": (0, 100),  # mg/dL (optimal <100)
    "Cholesterol in HDL [Mass/volume] in Serum or Plasma": (60, 80),       # mg/dL (ideal range)
    "Triglyceride [Mass/volume] in Serum or Plasma": (0, 150),             # mg/dL (normal <150)

    # Kidney function / kidney disease translation
    "Creatinine [Mass/volume] in Serum or Plasma": (0.6, 1.3),             # mg/dL (typical adult)
    "Urea nitrogen [Mass/volume] in Serum or Plasma": (7, 20),             # BUN mg/dL
    "Glomerular filtration rate/1.73 sq M.predicted [Volume Rate/Area] in Serum, Plasma or Blood by Creatinine-based formula (MDRD)": (90, 120),  # mL/min/1.73m^2 (normal ~>=90)
    "Albumin/Creatinine [Mass Ratio] in Urine": (0, 30),                   # mg/g (normal <30)

    # Inflammation / infection risk adjunct
    "C reactive protein [Mass/volume] in Serum or Plasma": (0, 10),         # mg/L (general reference; hs-CRP differs)
    "Leukocytes [#/volume] in Blood by Automated count": (4.0, 11.0),  # x10^3/uL
    "Erythrocytes [#/area] in Urine sediment by Microscopy high power field": (0, 2),  # RBC/HPF
    "Erythrocyte sedimentation rate by Westergren method": (0.0, 20.0),  # mm/hr (broad adult ref)
    
    # Anemia
    "Hemoglobin [Mass/volume] in Blood": (120.0, 175.0),  # g/L (broad adult range)
    "MCV [Entitic volume] by Automated count": (80.0, 100.0),  # fL
}


MEASUREMENT_NAMES_MAP = {
    "25-Hydroxyvitamin D3+25-Hydroxyvitamin D2 [Mass/volume] in Serum or Plasma": "vitamin_d_25oh_total",
    "Alanine aminotransferase [Enzymatic activity/volume] in Serum or Plasma": "alt",
    "Albumin [Mass/volume] in Serum or Plasma": "albumin_serum",
    "Albumin/Creatinine [Mass Ratio] in Urine": "uacr",
    "Alkaline phosphatase [Enzymatic activity/volume] in Serum or Plasma": "alp",
    "Anion gap 3 in Serum or Plasma": "anion_gap",
    "Aspartate aminotransferase [Enzymatic activity/volume] in Serum or Plasma": "ast",
    "Basophils [#/volume] in Blood": "basophils_abs",
    "Basophils [#/volume] in Blood by Automated count": "basophils_abs_auto",
    "Basophils/100 leukocytes in Blood": "basophils_pct",
    "Basophils/100 leukocytes in Blood by Automated count": "basophils_pct_auto",
    "Basophils/100 leukocytes in Blood by Manual count": "basophils_pct_manual",
    "Bicarbonate [Moles/volume] in Serum or Plasma": "bicarbonate",
    "Bilirubin.total [Mass/volume] in Serum or Plasma": "bilirubin_total",
    "Body height": "height",
    "Body mass index (BMI) [Ratio]": "bmi",
    "Body weight": "weight",
    "C reactive protein [Mass/volume] in Serum or Plasma": "crp",
    "Calcium [Mass/volume] in Serum or Plasma": "calcium",
    "Chloride [Moles/volume] in Serum or Plasma": "chloride",
    "Cholesterol [Mass/volume] in Serum or Plasma": "chol_total",
    "Cholesterol in HDL [Mass/volume] in Serum or Plasma": "hdl_c",
    "Cholesterol in LDL [Mass/volume] in Serum or Plasma by calculation": "ldl_c_calc",
    "Creatine kinase [Enzymatic activity/volume] in Serum or Plasma": "ck",
    "Creatinine [Mass/volume] in Serum or Plasma": "creatinine",
    "Diastolic blood pressure": "dbp",
    "Eosinophils [#/volume] in Blood by Automated count": "eosinophils_abs_auto",
    "Epithelial cells.squamous [#/area] in Urine sediment by Microscopy high power field": "urine_squamous_epi_hpf",
    "Erythrocyte distribution width [Entitic volume]": "rdw_sd",
    "Erythrocyte distribution width [Ratio] by Automated count": "rdw_cv_auto",
    "Erythrocyte sedimentation rate by Westergren method": "esr_westergren",
    "Erythrocytes [#/area] in Urine sediment by Microscopy high power field": "urine_rbc_hpf",
    "Erythrocytes [#/volume] in Blood by Automated count": "rbc_abs_auto",
    "Glomerular filtration rate/1.73 sq M.predicted [Volume Rate/Area] in Serum, Plasma or Blood by Creatinine-based formula (MDRD)": "egfr_mdrd",
    "Glucose [Mass/volume] in Serum or Plasma": "glucose",
    "Heart rate": "heart_rate",
    "Hematocrit [Volume Fraction] of Blood by Automated count": "hematocrit_auto",
    "Hemoglobin A1c/Hemoglobin.total in Blood": "hba1c",
    "Hemoglobin [Mass/volume] in Blood": "hemoglobin",
    "Immature granulocytes [#/volume] in Blood": "immature_granulocytes_abs",
    "Immature granulocytes [#/volume] in Blood by Automated count": "immature_granulocytes_abs_auto",
    "Immature granulocytes/100 leukocytes in Blood": "immature_granulocytes_pct",
    "Immature granulocytes/100 leukocytes in Blood by Automated count": "immature_granulocytes_pct_auto",
    "Leukocytes [#/area] in Urine sediment by Microscopy high power field": "urine_wbc_hpf",
    "Leukocytes [#/volume] in Blood by Automated count": "wbc_abs_auto",
    "Lymphocytes [#/volume] in Blood by Automated count": "lymphocytes_abs_auto",
    "MCH [Entitic mass] by Automated count": "mch_auto",
    "MCHC [Mass/volume] by Automated count": "mchc_auto",
    "MCV [Entitic volume] by Automated count": "mcv_auto",
    "Magnesium [Mass/volume] in Serum or Plasma": "magnesium",
    "Monocytes [#/volume] in Blood by Automated count": "monocytes_abs_auto",
    "Natriuretic peptide B [Mass/volume] in Serum or Plasma": "bnp",
    "Natriuretic peptide.B prohormone N-Terminal [Mass/volume] in Serum or Plasma": "ntprobnp",
    "Neutrophils [#/volume] in Blood by Automated count": "neutrophils_abs_auto",
    "Nucleated erythrocytes [#/volume] in Blood by Automated count": "nrbc_abs_auto",
    "Nucleated erythrocytes/100 leukocytes [Ratio] in Blood": "nrbc_pct",
    "Nucleated erythrocytes/100 leukocytes [Ratio] in Blood by Automated count": "nrbc_pct_auto",
    "Phosphate [Mass/volume] in Serum or Plasma": "phosphate",
    "Platelet mean volume [Entitic volume] in Blood by Automated count": "mpv_auto",
    "Platelets [#/volume] in Blood by Automated count": "platelets_abs_auto",
    "Potassium [Moles/volume] in Serum or Plasma": "potassium",
    "Protein [Mass/volume] in Serum or Plasma": "protein_total_serum",
    "Protein [Mass/volume] in Urine by Test strip": "urine_protein_dipstick",
    "Sodium [Moles/volume] in Serum or Plasma": "sodium",
    "Systolic blood pressure": "sbp",
    "Thyroxine (T4) free [Mass/volume] in Serum or Plasma": "free_t4",
    "Triglyceride [Mass/volume] in Serum or Plasma": "triglycerides",
    "Troponin I.cardiac [Mass/volume] in Serum or Plasma": "troponin_i",
    "Urea nitrogen [Mass/volume] in Serum or Plasma": "bun",
}





import os
from functools import lru_cache
from importlib.resources import as_file, files

import pandas as pd
import requests

from rwe.clients.hgnc import get_gene_info


_CONSTRAINT_RESOURCE = "gnomad.v4.1.constraint_metrics.tsv"
_PLOF_SUMMARY_RESOURCE = "Combined_plof_summary_Gnomad_GeneBass_Guo_2026.csv"
_CONSTRAINT_URL = "https://storage.googleapis.com/gcp-public-data--gnomad/release/4.1/constraint/gnomad.v4.1.constraint_metrics.tsv"
_HGNC_RESOURCE = "gene_with_protein_product.txt"


def _normalize_gene(gene: str) -> str:
    return gene.strip().upper()


def _default_constraint_path() -> str:
    return os.path.expanduser(f"~/.cache/rwe/gnomad/{_CONSTRAINT_RESOURCE}")


def _default_hgnc_path() -> str:
    return os.path.abspath(_HGNC_RESOURCE)


def _download_constraint_table(constraint_path: str) -> None:
    os.makedirs(os.path.dirname(constraint_path), exist_ok=True)
    resp = requests.get(_CONSTRAINT_URL, stream=True, timeout=120)
    resp.raise_for_status()
    with open(constraint_path, "wb") as f:
        for chunk in resp.iter_content(chunk_size=1024 * 1024):
            if chunk:
                f.write(chunk)
    return

@lru_cache(maxsize=4)
def _load_gnomad_constraint_table(constraint_path: str) -> pd.DataFrame:
    if not os.path.exists(constraint_path):
        _download_constraint_table(constraint_path)

    df = pd.read_csv(
        constraint_path,
        sep="\t",
        usecols=["gene", "lof.oe_ci.upper", "canonical", "mane_select", "level"],
        low_memory=False,
    )
    df["gene"] = df["gene"].astype(str).str.upper()
    return df


@lru_cache(maxsize=1)
def _load_gnomad_plof_summary() -> pd.DataFrame:
    resource = files("rwe") / "assets" / _PLOF_SUMMARY_RESOURCE
    with as_file(resource) as path:
        df = pd.read_csv(
            path,
            usecols=["Gene", "Homozygote_Count", "Heterozygote_Count", "source"],
        )
    df["Gene"] = df["Gene"].astype(str).str.upper()
    df["source"] = df["source"].astype(str).str.lower()
    return df.loc[df["source"] == "gnomad"].copy()


def _select_best_constraint_row(gene_df: pd.DataFrame) -> pd.Series:
    ranked = gene_df.assign(
        _gene_level=gene_df["level"].isna(),
        _canonical=gene_df["canonical"].astype(str).str.lower() == "true",
        _mane_select=gene_df["mane_select"].astype(str).str.lower() == "true",
    ).sort_values(
        by=["_gene_level", "_canonical", "_mane_select"],
        ascending=[False, False, False],
    )
    return ranked.iloc[0]


def get_gnomad_constraint_info(
    gene: str,
    constraint_path: str | None = None,
    hgnc_path: str | None = None,
) -> dict:
    """
    Return gnomAD constraint and carrier count information for a gene.

    The lookup is performed against:
    - a downloaded local copy of `gnomad.v4.1.constraint_metrics.tsv` for LOEUF (`lof.oe_ci.upper`)
    - `Combined_plof_summary_Gnomad_GeneBass_Guo_2026.csv` for gnomAD carrier counts
    """
    hgnc_path = hgnc_path or _default_hgnc_path()
    approved_symbol, _ensembl_id, _entrez_id = get_gene_info(gene, hgnc_path)
    normalized_gene = _normalize_gene(approved_symbol)
    constraint_path = constraint_path or _default_constraint_path()

    constraint_df = _load_gnomad_constraint_table(constraint_path)
    constraint_hits = constraint_df.loc[constraint_df["gene"] == normalized_gene]
    if constraint_hits.empty:
        loeuf = None
    else:
        loeuf = _select_best_constraint_row(constraint_hits)["lof.oe_ci.upper"]
        loeuf = None if pd.isna(loeuf) else float(loeuf)

    plof_df = _load_gnomad_plof_summary()
    plof_hits = plof_df.loc[plof_df["Gene"] == normalized_gene]
    if plof_hits.empty:
        nhomz = None
        nhetz = None
    else:
        row = plof_hits.iloc[0]
        nhomz = None if pd.isna(row["Homozygote_Count"]) else int(row["Homozygote_Count"])
        nhetz = None if pd.isna(row["Heterozygote_Count"]) else int(row["Heterozygote_Count"])

    return {
        "input_gene": gene,
        "gene": normalized_gene,
        "loeuf": loeuf,
        "nhetz": nhetz,
        "nhomz": nhomz,
    }

if __name__ == "__main__":
    gene = "ANGPTL3"
    info = get_gnomad_constraint_info(
        gene, 
        constraint_path="/home/dbanerjee/work/rwe/data/rwe_reports/gnomad.v4.1.constraint_metrics.tsv",
        hgnc_path="/home/dbanerjee/work/rwe/data/rwe_reports/gene_with_protein_product.txt"
        )
    print(info)



import pandas as pd
# plotting imports 
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42
# matplotlib.rcParams['font.sans-serif'] = "Arial" # missing fonts:: https://alexanderlabwhoi.github.io/post/2021-03-missingfont/
# Then, "ALWAYS use sans-serif fonts"
matplotlib.rcParams['font.family'] = "sans-serif"
matplotlib.rcParams.update({'font.size': 6, 'axes.linewidth': 1, 'xtick.major.width': 1, 'xtick.major.size': 5, 'ytick.major.width': 1, 'ytick.major.size': 5})


def percent_bar_with_n(df, group_col, cat_col, title=None, top_n=None,
                       control_label="Controls", case_label="Cases",
                       figsize=(8, 6),
                       ax=None, show=True,
                       palette=None):
    
    d = df[[group_col, cat_col]].copy()
    d[group_col] = (d[group_col]
                    .fillna("Unknown")
                    .astype(str)
                    .replace({"": "Unknown", "nan": "Unknown", "None": "Unknown"}))

    d[cat_col] = (d[cat_col]
                  .fillna("Unknown")
                  .astype(str)
                  .replace({"": "Unknown", "nan": "Unknown", "None": "Unknown"}))
    # Optional: collapse rare categories (global frequency)
    if top_n is not None:
        top = d[cat_col].value_counts().head(top_n).index
        d[cat_col] = d[cat_col].where(d[cat_col].isin(top), "Other")
    # counts + within-group percent
    ct = d.groupby([group_col, cat_col]).size().reset_index(name="n")
    ct["pct"] = ct["n"] / ct.groupby(group_col)["n"].transform("sum") * 100
    # enforce hue order
    present = ct[group_col].unique().tolist()
    hue_order = [control_label, case_label]
    if control_label not in present or case_label not in present:
        hue_order = sorted(present)
    # category order: sort by CONTROL pct desc
    ctrl = ct[ct[group_col] == hue_order[0]]
    if ctrl.empty:
        cat_order = d[cat_col].value_counts().index.tolist()
    else:
        cat_order = ctrl.sort_values("pct", ascending=False)[cat_col].tolist()
    # append categories missing from controls (keep at end)
    cat_order += [c for c in d[cat_col].unique().tolist() if c not in cat_order]
    # FORCE categorical order BEFORE plotting
    ct[cat_col] = pd.Categorical(ct[cat_col], categories=cat_order, ordered=True)
    ct[group_col] = pd.Categorical(ct[group_col], categories=hue_order, ordered=True)
    ct = ct.sort_values([cat_col, group_col])
    # plotting
    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = plt.gca()
    else:
        fig = ax.figure

    sns.barplot(
        data=ct,
        x=cat_col, y="pct",
        hue=group_col,
        gap=0.1,
        order=cat_order,
        hue_order=hue_order,
        estimator=sum,
        errorbar=None,
        ax=ax,
        palette=palette
    )

    ax.set_title(title or f"{cat_col} (% within group)")
    ax.set_xlabel("")
    ax.set_ylabel("Percent (%)")
    ax.tick_params(axis="x", rotation=25)
    for lbl in ax.get_xticklabels():
        lbl.set_ha("right")

    # annotate n
    for container, grp in zip(ax.containers, hue_order):
        ns = (ct[ct[group_col] == grp]
              .set_index(cat_col)
              .reindex(cat_order)["n"]
              .fillna(0)
              .astype(int)
              .tolist())

        labels = [f"{n}" if rect.get_height() > 0 else "" for rect, n in zip(container, ns)]
        if hasattr(ax, "bar_label"):
            ax.bar_label(container, labels=labels, padding=3, rotation=90)
        else:
            for rect, lab in zip(container, labels):
                if not lab:
                    continue
                h = rect.get_height()
                ax.text(rect.get_x() + rect.get_width()/2, h, lab,
                        ha="center", va="bottom", rotation=90)

    ax.spines[['top', 'right']].set_visible(False)

    if ax.get_legend() is not None:
        ax.get_legend().set_title("")

    if show and ax is not None and ax.figure is not None and ax.figure.get_axes()[0] == ax:
        plt.tight_layout()
        plt.show()

    return fig, ax


def demographics_plot(person_df,
                        group_col="group",
                        control_label="Controls",
                        case_label="Cases",
                        sex_col="sex_at_birth",
                        ethnicity_col="ethnicity",
                        ancestry_col="ancestry_pred_other",
                        age_col="age_at_cdr",
                        top_n_sex=2,
                        top_n_ethnicity=2,
                        top_n_ancestry=None,
                        figsize=(4, 5),
                        palette=None,
                        share_legend=True,
                        savepath=None,
                        dpi=300):

    if palette is None:
        palette = {
            control_label: "#2F6690",  # BLUE
            case_label:    "#D1495B",  # RED
        }

    fig = plt.figure(figsize=figsize)
    gs = fig.add_gridspec(nrows=2, ncols=4, wspace=1, hspace=0.5)

    ax_age = fig.add_subplot(gs[0, 0:1])
    ax_ancestry      = fig.add_subplot(gs[0, 1:4])
    ax_sex      = fig.add_subplot(gs[1, 0:2])
    ax_eth      = fig.add_subplot(gs[1, 2:4])

    # Top row
    percent_bar_with_n(
        person_df, group_col=group_col, cat_col=ancestry_col,
        title="Ancestry", top_n=top_n_ancestry,
        control_label=control_label, case_label=case_label,
        ax=ax_ancestry, show=False, palette=palette
    )

    sns.violinplot(
        data=person_df, y=age_col, hue=group_col,
        ax=ax_age, palette=palette,# cut=0, inner="quartile"
    )
    ax_age.set_title("Age")
    ax_age.set_xlabel("")
    ax_age.set_ylabel("Age")
    ax_age.spines[['top', 'right']].set_visible(False)
    if ax_age.get_legend() is not None:
        ax_age.get_legend().set_title("")

    # Bottom row
    percent_bar_with_n(
        person_df, group_col=group_col, cat_col=sex_col,
        title="Sex at Birth", top_n=top_n_sex,
        control_label=control_label, case_label=case_label,
        ax=ax_sex, show=False, palette=palette
    )

    percent_bar_with_n(
        person_df, group_col=group_col, cat_col=ethnicity_col,
        title="Ethnicity", top_n=top_n_ethnicity,
        control_label=control_label, case_label=case_label,
        ax=ax_eth, show=False, palette=palette
    )

    axes = [ax_ancestry, ax_age, ax_sex, ax_eth]

    # One shared legend
    if share_legend:
        handles_labels = None
        for ax in axes:
            leg = ax.get_legend()
            if leg is not None and handles_labels is None:
                handles_labels = ax.get_legend_handles_labels()
            if leg is not None:
                leg.remove()

        if handles_labels is not None:
            handles, labels = handles_labels
            fig.legend(handles, labels, loc="upper center",
                       ncol=len(labels), frameon=False, bbox_to_anchor=(0.5, 1.0))

    fig.tight_layout()

    if savepath is not None:
        fig.savefig(savepath, dpi=dpi, bbox_inches="tight")

    return fig, {
        "ancestry": ax_ancestry,
        "age": ax_age,
        "sex_at_birth": ax_sex,
        "ethnicity": ax_eth
    }



import pandas as pd
import os
from tqdm import tqdm
import numpy as np
import ast
# plotting imports 
import textwrap
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator
import matplotlib.patches as mpatches
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42
# matplotlib.rcParams['font.sans-serif'] = "Arial" # missing fonts:: https://alexanderlabwhoi.github.io/post/2021-03-missingfont/
# Then, "ALWAYS use sans-serif fonts"
matplotlib.rcParams['font.family'] = "sans-serif"
matplotlib.rcParams.update({'font.size': 6, 'axes.linewidth': 1, 'xtick.major.width': 1, 'xtick.major.size': 5, 'ytick.major.width': 1, 'ytick.major.size': 5})
from matplotlib.colors import TwoSlopeNorm
from matplotlib.backends.backend_pdf import PdfPages


def percent_bar_with_n(
    df, group_col, cat_col, cat_order=None, title=None, top_n=None,
    control_label="Controls", case_label="Cases",
    figsize=(8, 6), ax=None, show=True, palette=None
    ):

    d = df[[group_col, cat_col]].copy()
    d[group_col] = (d[group_col]
                    .fillna("Unknown")
                    .astype(str)
                    .replace({"": "Unknown", "nan": "Unknown", "None": "Unknown"}))

    d[cat_col] = (d[cat_col]
                  .fillna("Unknown")
                  .astype(str)
                  .replace({"": "Unknown", "nan": "Unknown", "None": "Unknown"}))
    # Optional: collapse rare categories (global frequency)
    if top_n is not None:
        top = d[cat_col].value_counts().head(top_n).index
        d[cat_col] = d[cat_col].where(d[cat_col].isin(top), "Other")
    # counts + within-group percent
    ct = d.groupby([group_col, cat_col]).size().reset_index(name="n")
    ct["pct"] = ct["n"] / ct.groupby(group_col)["n"].transform("sum") * 100
    # enforce hue order
    present = ct[group_col].unique().tolist()
    hue_order = [control_label, case_label]
    if control_label not in present or case_label not in present:
        hue_order = sorted(present)
    # category order: sort by given categories if present else by control pct desc
    if cat_order is not None:
        present_cats = d[cat_col].unique().tolist()
        cat_order = [c for c in cat_order if c in present_cats]          # keep only those that exist
        cat_order += [c for c in present_cats if c not in cat_order]     # append anything else (Unknown/Other/etc.)
    else:
        ctrl = ct[ct[group_col] == hue_order[0]]
        if ctrl.empty:
            cat_order = d[cat_col].value_counts().index.tolist()
        else:
            cat_order = ctrl.sort_values("pct", ascending=False)[cat_col].tolist()
        cat_order += [c for c in d[cat_col].unique().tolist() if c not in cat_order]
    # FORCE categorical order BEFORE plotting
    ct[cat_col] = pd.Categorical(ct[cat_col], categories=cat_order, ordered=True)
    ct[group_col] = pd.Categorical(ct[group_col], categories=hue_order, ordered=True)
    ct = ct.sort_values([cat_col, group_col])
    # plotting
    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = plt.gca()
    else:
        fig = ax.figure

    sns.barplot(
        data=ct,
        x=cat_col, y="pct",
        hue=group_col,
        gap=0.1,
        order=cat_order,
        hue_order=hue_order,
        estimator=sum,
        errorbar=None,
        ax=ax,
        palette=palette
    )

    ax.set_title(title or f"{cat_col} (% within group)", pad=20)
    ax.set_xlabel("")
    ax.set_ylabel("Percent (%)")
    ax.tick_params(axis="x", rotation=25)
    for lbl in ax.get_xticklabels():
        lbl.set_ha("right")

    # annotate n
    for container, grp in zip(ax.containers, hue_order):
        ns = (ct[ct[group_col] == grp]
              .set_index(cat_col)
              .reindex(cat_order)["n"]
              .fillna(0)
              .astype(int)
              .tolist())

        labels = [f"{n}" if rect.get_height() > 0 else "" for rect, n in zip(container, ns)]
        if hasattr(ax, "bar_label"):
            ax.bar_label(container, labels=labels, padding=3, rotation=90)
        else:
            for rect, lab in zip(container, labels):
                if not lab:
                    continue
                h = rect.get_height()
                ax.text(rect.get_x() + rect.get_width()/2, h, lab,
                        ha="center", va="bottom", rotation=90)

    ax.spines[['top', 'right']].set_visible(False)

    if ax.get_legend() is not None:
        ax.get_legend().set_title("")

    if show and ax is not None and ax.figure is not None and ax.figure.get_axes()[0] == ax:
        plt.tight_layout()
        plt.show()

    return fig, ax

# Create a single survey figure
def plot_survey_questions(
    df,
    surveys=None,
    group_col="cases",
    cat_col="answer_category",      # or "answer"
    survey_col="survey",
    question_col="question",
    control_label="Controls",
    case_label="Cases",
    top_n=10,
    figsize=(8, 4),
    palette=None):

    if palette is None:
        palette = {
            control_label: "#2F6690",  # BLUE
            case_label:    "#D1495B",  # RED
        }

    d = df.copy()

    # optional: restrict to chosen surveys
    if surveys is not None:
        d = d[d[survey_col].isin(surveys)].copy()

    # map True/False -> nice labels (so your function's hue_order works)
    d[group_col] = np.where(d[group_col].astype(bool), case_label, control_label)

    groups = list(d.groupby([survey_col, question_col], dropna=False))
    n = len(groups)
    ncols = 2
    nrows = int(np.ceil(n / ncols))
    
    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)
    axes = axes.ravel()
    handles = labels = None

    for i, ((svy, q), sub) in enumerate(groups):
        percent_bar_with_n(
            sub,
            group_col=group_col,
            cat_col=cat_col,
            cat_order=["Poor","Fair","Good","Very Good","Excellent"],
            title=f"{q}",
            top_n=top_n,
            control_label=control_label,
            case_label=case_label,
            ax=axes[i],
            show=False,          # << key
            palette=palette
        )

        # remove per-axis legends; we'll add one global legend
        if handles is None:
            handles, labels = axes[i].get_legend_handles_labels()
        leg = axes[i].get_legend()
        if leg is not None:
            leg.remove()

    # hide unused axes
    for j in range(n, len(axes)):
        axes[j].axis("off")

    # global legend from first axis handles
    if handles:
        fig.legend(handles, labels, loc="upper right", frameon=False)

    fig.tight_layout()
    return fig, axes



disease_to_keywords = {
    "obesity": [
        "obes", "overweight", "BMI", "body mass index", "adipos", "central obesity", "waist",
        "hip circ", "waist to hip", "waist-to-hip", "WHR", "body weight", "body fat", "fat mass",
        "percent body fat", "waist circumference", "hip circumference",
    ],
    "type_2_diabetes": [
        "type 2", "type-2", "T2D", "diabet", "noninsulin", "NIDDM", "hypergly", 
        "glucose", "fasting glucose", "fasting plasma glucose", "HbA1c",
        "A1c", "glycated haemoglobin", "OGTT", "oral glucose tolerance",
        "2-hour glucose", "2 hour glucose", "insulin", "HOMA", "insulin resistance",
    ],
    "diabetic_kidney_disease": [
        "diabet", "DKD", "diabetic nephrop", "diabetic kidney",
        "diabetic CKD", "chronic kidney", "CKD", "nephrop", "nephrit",
        "renal", "eGFR", "glomerular", "creatinine", "serum creatinine",
        "albuminuria", "microalbumin", "macroalbumin", "uACR",
        "albumin-to-creatinine", "albumin to creatinine", "proteinuria", "albumin",
        "dialysis", "urine albumin", "urine protein",
    ],
    "dyslipidemia": [
        "dyslipid", "hyperlipid", "hypercholesterol", "cholesterol", "ldl", "ldl-c",
        "ldl c", "hdl", "hdl-c", "hdl c", "triglycer", "apolipoprotein", "apob",
        "apoa1", "apoa-1", "lipoprotein", "lpa", "non-hdl",
        "non hdl", "vldl", "remnant",
    ],
    "cold_agglutinin_disease": [
        "autoimmune hemolytic", "hemolytic anemia", "raynaud", "hemoglobinuria",
        "bilirubin", "hemoglobin", "haemoglobin", "heart rate", "hematocrit", "erythrocyte", "rbc",
        "wbc", "hematuria", "mcv", "mch", "mchc", "corpuscular", "red blood", "white blood", "reticulocyte"
    ],
    "long_qt_syndrome": [
            "long qt", "lqts", "qt prolong", "prolonged qt", "qtc", "qt interval", "torsad",
            "channelopath", "romano", "jervell", "lange",
            "arrhythm", "ventricular tachy", "ventricular fibrill", "flutter",
            "supraventricular tachy", "paroxysmal", "polymorphic",
            "syncope", "collapse", "palpit", "cardiac arrest", "sudden",
            "ecg", "ekg", "electrocard", "qrs", "loop recorder",
            "potassium", "magnesium", "calcium", "sodium", "bicarbonate", "chloride", "anion gap",
            "heart rate", "troponin", "natriuretic",
        ],
    "hypertrophic_cardiomyopathy": [
        "cardiomyo", "hocm", "hcm", "cardiomegaly", "heart failure", "pulmonary oed", "pulmonary od",
        "diastolic", "systolic", "hfpef", "reduced ef", "preserved ef",
        "atrial fibrill", "flutter", "arrhythm", "ventricular tachy", "ventricular fibrill",
        "syncope", "palpit", "cardiac arrest", "echoc", "ecg", "ekg", "electrocard", "qrs", 
        "qt interval", "qtc", "cardioverter defibr", "defibrillator", "pacemaker", "loop recorder",
        "ablation of atr", "ablation of pulmonary", "ablation of atrioventricular", "conducting system",
        "heart rate", "troponin", "natriuretic", "creatine kinase", "heart murmur", "heart valve problem",
    ],
    "metabolic_syndrome": [
        "metabolic", "syndrome x", "glucose", "HbA1c", "haemoglobin", "hemoglobin",
        "insulin", "HOMA", "triglycer", "hdl","blood pressure", "hypertens",
        "waist", "central obes", "visceral", "BMI", "body mass index", "obes", "overweight", "adipos",
        "urate", "uric acid", "c-reactive protein", "CRP",
        "alanine aminotransf", "ALT", "aspartate aminotransf", "AST",
        "gamma glutamyl", "GGT", "alkaline phosphat",
        "liver PDFF", "fatty liver", "NAFLD", "NASH", "cholesterol", "ldl", "apolipoprotein",
    ],
    "angelman_syndrome": [
        "angelman", "happy puppet", "chromosome 15", "UBE3A", "15q11", "15q12", "chromosomal anomal",
        "epilep", "seizure", "convuls", "status epilepticus",
        "intellectual disab", "mental retard", "developmental delay",
        "speech", "language disorder", "atax", "cerebell", "gait", "movement disorder",
        "tremor", "nystagmus", "irregular eye movement",
        "scoliosis", "microceph", "sleep", "melatonin",
        "cognitive", "fluid intelligence",
    ],
    "hemophilia": [
        "hemophil", "haemophil", "coagul", "clotting", "bleeding", "hemorrhagic",
        "von willebrand", "vwf", "aptt", "prothrombin", "INR", "fibrinogen", "d-dimer", 
        "purpura", "hemarthro", "joint bleed", "platelet"
    ],
    "essential_thrombocythemia": [
        "thrombocyt", "myeloprolif", "polycyth", "myelofibro", "myelodysplas",
        "platelet", "leukocyte", "blood cell", "wbc",
        "erythrocyte", "rbc", "haemoglobin", "hemoglobin", "hematocrit", "haematocrit",
        "neutroph", "basoph", "eosinoph", "lymphocyte", "monocyte", "reticulocyte",
        "LDH", "lactate dehydrogen", "ferritin", "iron",
        "urate", "uric acid", "haptoglobin", "spleen", "splenomeg"
    ],
    "aortic_valve_stenosis": [
        "aortic", "valve", "stenos", "heart", "endocard",
        "bicuspid", "calcif", "mitral", "tricuspid", "rheumatic heart",
        "echoc", "ejection fraction", "LV end", "cardiac",
        "end systolic", "end diastolic", "cardiomegaly",
        "atrial fibrill", "arrhythm", "ecg", "ekg", "electrocard",
        "natriuretic", "NTproBNP", "bnp", "troponin",
        "heart rate", "pulse rate", "pulse wave",
        "blood pressure", "systolic", "diastolic",
        "heart failure", "aortic aneur", "aortic dissec",
        "syncope", "dyspn", "calcium", "phosphat", "alkaline phosphat", "creatinine",
    ],
    "nash": [
        "nash", "steatohepat", "nonalcohol", "nafld",         
        "fatty liver", "steatosis", "fibrosis", "cirrhosis",
        ],
    "sepsis": [
        "sepsis", "septic", "bacteremia", "sirs",
    ],
    "kidney_disease": [
        # Core disease stems (catch the long tail of subtypes)
        "kidney", "renal", "nephro", "nephri", "nephrect",
        "glomerul", "tubul", "pyelon", "polycyst",
        # Abbreviations that may appear without "kidney"/"renal"
        "CKD", "ESRD", "eGFR",
        # Quantitative biomarkers
        "creatinine", "cystatin", "albumin", "proteinur",
        "uACR", "urea", "urate", "uric acid",
        # Procedures (catches haemodialysis/hemodialysis too)
        "dialys",
        # Secondary complications
        "hyperparathyroid",
    ],
    "menopause": [
        # Core disease terms
        "menopaus", "climacter", "perimenopaus",
        # Ovarian status / surgical menopause
        "oophorect", "ovarian fail", "ovarian insuff",
        "ovarian dysfunc", "amenorrh",
        "age at menopause",
        # VMS symptoms
        "vasomotor", "hot flash", "hot flush",
        "flush", "night sweat", "sweat",
        "hyperhidr", "eccrine",
        # Hormonal biomarkers
        "oestradiol", "estradiol", "SHBG", "testosterone",
        "FSH", "progesterone",
        # HRT
        "hormone replace", "HRT",
    ],
    "noonan_syndrome": [
        # Cardiac
        "pulmon", "cardiomyop", "septal defect",
        # Growth
        "height", "stature", "IGF", "growth hormone", "failure to thrive",
        # Hematologic
        "platelet", "thrombocyt", "coagul", "bleeding", "factor VIII", "myeloid",
        # Lymphatic
        "lymphedem", "lymphoedem",
        # Skeletal
        "pectus", "scoliosis",
        # Craniofacial / ocular
        "ptosis", "strabism", "hearing",
        # Abdominal
        "spleen", "hepatomeg",
        # Genitourinary
        "cryptorch", "undescend",
    ],
}

arrowhead_disease_to_keywords = {
    # Cardiometabolic
    "familial_chylomicronemia_syndrome": [
        "familial chylomicronemia", "chylomicronemia",
        "fcs", "chylomicron", "lipoprotein lipase", "lpl",
        "triglycer", "hypertriglycer", "tg", "pancreatitis",
    ],
    "severe_hypertriglyceridemia": [
        "severe hypertriglycer", "hypertriglycer", "triglycer",
        "triglyceride", "tg", "pancreatitis",
    ],
    "homozygous_familial_hypercholesterolemia": [
        "homozygous familial hypercholesterol",
        "homozygous familial hypercholesterolemia",
        "homozygous familial hypercholesterolaemia",
        "hofh",
        "familial hypercholesterol",
        "familial hypercholesterolemia",
        "familial hypercholesterolaemia",
        "fh",
        "ldl receptor",
        "ldlr",
        # include lipid measurements because they dominate in your exports
        "cholesterol",
        "ldl",
        "ldl-c",
        "ldl c",
        "hypercholesterol",
        "hypercholesterolaemia",
    ],
    "cardiovascular_disease": [
        # disease/clinical terms (broad to hit the cardio phenos present in your exports)
        "cardiovascular",
        "cardio",
        "heart",
        "cardiac",
        "vascular",
        "artery",
        "arterial",
        "coronary",
        "coronary artery",
        "coronary heart",
        "ischemic heart",
        "myocard",
        "myocardial infarction",
        "heart attack",
        "stroke",
        "ischemic stroke",
        "angina",
        "atheroscleros",
        "arterioscleros",
        "peripheral artery",
        "revascular",
        "stent",
        "bypass",
        # major related phenotypes in these exports
        "hypertens",
        "blood pressure",
        "systolic",
        "diastolic",
        "heart failure",
        "congestive heart",
        "cardiomyopathy",
        "atrial fibrillation",
        "arrhythm",
        "tachycard",
        "bradycard",
        # clotting / thrombotic events frequently appear in phenotype catalogs
        "thromb",
        "embol",
        "venous thromboembol",
        "pulmonary embol",
        "dvt",
    ],
    "obesity": [
        "obes",
        "overweight",
        "body mass index",
        "bmi",
        "adipos",
        "central obesity",
        "waist",
        "hip",
        "waist to hip",
        "waist-to-hip",
        "whr",
        # measurements/body comp that show up in continuous trait catalogs
        "body weight",
        "#weight",
        "weight gain",
        "body fat",
        "fat mass",
        "percent body fat",
        "lean mass",
        "body composition",
        "waist circumference",
        "hip circumference",
    ],
    "mash": [
        "mash",
        "metabolic dysfunction-associated steatohepatitis",
        "metabolic associated steatohepatitis",
        "nash",
        "nonalcoholic steatohepatitis",
        "non-alcoholic steatohepatitis",
        "steatohepatitis",
        "fatty liver",
        "fatty liver disease",
        "hepatic steatosis",
        "steatosis",
        "nafld",
        "nonalcoholic fatty liver",
        "non-alcoholic fatty liver",
        "liver fibrosis",
        "fibrosis",
        "cirrhosis",
        # common liver measurements in these exports
        "alt",
        "ast",
        "ggt",
        "alp",
        "bilirubin",
        "inr",
        "platelet",
        "liver fat",
        "hepatic fat",
        "fibroscan",
        "elastography",
        "transaminase",
        "liver enzyme",
    ],

    # CNS
    "spinocerebellar_ataxia_2": [
        "spinocerebellar ataxia",
        "sca2",
        "sca 2",
        "ataxia 2",
        "ataxia",
        "cerebellar",
        "spinocerebellar",
        "atxn2",
    ],

    # Pulmonary
    "inflammatory_pulmonary_diseases": [
        "inflammatory pulmonary",
        "pulmonary inflammation",
        "lung inflammation",
        "inflammatory lung",
        "pneumonitis",
        "pulmonary",
        "lung",
        "respirat",
        "interstitial lung disease",
        "ild",
        "copd",
        "asthma",
        "bronchitis",
        "emphysema",
        "pneumonia",
        "pneumon",
        "dyspnea",
        "dyspnoea",
        "shortness of breath",
        "wheeze",
        # pulmonary function measurements that appear as continuous traits
        "fev1",
        "fvc",
        "dlco",
        "spirom",
        # gene/program hints that can appear in some catalogs
        "rage",
        "ager",
        # optional but common pulmonary phenotype bucket
        "sleep apnea",
        "sleep apnoea",
    ],
    "idiopathic_pulmonary_fibrosis": [
        "idiopathic pulmonary fibrosis",
        "ipf",
        "pulmonary fibrosis",
        "lung fibrosis",
        "fibrosis of lung",
        "fibrosing",
        "fibrosis",
        "fibrotic",
        "usual interstitial",
        "interstitial pneumon",
        "ild",
        "pulmonary",
        "lung",
        # program hint
        "mmp7",
    ],

    # Liver
    "alpha_1_liver_disease": [
        "alpha-1",
        "alpha 1",
        "alpha1",
        "alpha-1 antitrypsin",
        "alpha 1 antitrypsin",
        "antitrypsin",
        "a1at",
        "a1atd",
        "aat deficiency",
        "antitrypsin deficiency",
        # liver involvement + common liver labs
        "liver",
        "hepatic",
        "liver fibrosis",
        "cirrhosis",
        "alt",
        "ast",
        "ggt",
        "alp",
        "bilirubin",
        "inr",
        "platelet",
    ],
    "hepatitis_b": [
        "hepatitis b",
        "hbv",
        "chronic hepatitis b",
        "hep b",
        "hepatitis b virus",
        # virology measurements (often present as trait strings)
        "hbsag",
        "hbeag",
        "hbv dna",
        "viral load",
        # common liver injury labs
        "alt",
        "ast",
        "bilirubin",
        # broad safety net
        "hepatitis",
        "viral hepatitis",
    ],

    # Muscular
    "fshd": [
        "fshd",
        "facioscapulohumeral",
        "facioscapulohumeral muscular dystrophy",
        "muscular dystrophy",
        "dux4",
    ],
    "myotonic_dystrophy_type_1": [
        "myotonic dystrophy",
        "dm1",
        "dm 1",
        "dystrophia myotonica",
        "steinert",
        "steinert disease",
        "myotonia",
        "dmpk",
    ],

    # Complement Mediated Disease (ARO-C3 / ARO-CFB on pipeline page)
    "complement_mediated_disease": [
        "complement mediated",
        "complement-mediated",
        "complement",
        "alternative complement",
        "complement pathway",
        "c3",
        "complement component 3",
        "cfb",
        "factor b",
        "complement factor b",
        # common complement-mediated clinical phenos seen in catalogs
        "iga nephrop",
        "iga nephropathy",
        "c3 glomerulopathy",
        "glomerulonephritis",
        "gn",
        "ahus",
        "atypical hemolytic uremic",
        "pnh",
        "paroxysmal nocturnal hemoglobinuria",
        # kidney measurements (these dominate in your exports for this bucket)
        "egfr",
        "glomerular filtration",
        "creatinine",
        "cystatin",
        "albuminuria",
        "microalbumin",
        "proteinuria",
        "uacr",
        "albumin-to-creatinine",
        "urine albumin",
        "urine protein",
    ],
}



import os
import random
import re
from dataclasses import dataclass
from functools import lru_cache
from importlib.resources import as_file, files

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from docx.shared import Inches
from playwright.sync_api import Playwright, sync_playwright

import rwe.utils.helpers as uth
from rwe.clients.hgnc import get_gene_info
from rwe.plots.clinical import manhattan

_PLOF_SUMMARY_RESOURCE = "Combined_plof_summary_Gnomad_GeneBass_Guo_2026.csv"

_GENEBASS_CATEGORY_SHORT = {
    "Blood, blood-forming organs and certain immune disorders": "Blood/Immune",
    "Certain conditions originating in the perinatal period": "Perinatal",
    "Certain infectious and parasitic diseases": "Infectious",
    "Circulatory system disorders": "Circulatory",
    "Congenital disruptions and chromosomal abnormalities": "Congenital",
    "Digestive system disorders": "Digestive",
    "Ear and mastoid process disorders": "Ear/Mastoid",
    "Endocrine, nutritional and metabolic diseases": "Endocrine/Metabolic",
    "Eye and adnexa disorders": "Eye",
    "Genitourinary system disorders": "Genitourinary",
    "Mental and behavioural disorders": "Mental/Behavioral",
    "Musculoskeletal system and connective tissue disorders": "Musculoskeletal",
    "Nervous system disorders": "Nervous System",
    "Pregnancy, childbirth and the puerperium": "Pregnancy/Childbirth",
    "Respiratory system disorders": "Respiratory",
    "Skin and subcutaneous tissue disorders": "Skin",
}


@dataclass
class GeneBassClient:
    gene: str
    phewas_dir: str
    phewas_filename: str = ""

    def __post_init__(self) -> None:
        self.input_gene = self.gene
        self._approved_symbol = None
        self._ensembl_id = None

    @staticmethod
    @lru_cache(maxsize=1)
    def load_plof_summary() -> pd.DataFrame:
        resource = files("rwe") / "assets" / _PLOF_SUMMARY_RESOURCE
        with as_file(resource) as path:
            df = pd.read_csv(
                path,
                usecols=["Gene", "Homozygote_Count", "Heterozygote_Count", "source"],
            )
        df["Gene"] = df["Gene"].astype(str).str.upper()
        df["source"] = df["source"].astype(str)
        return df.loc[df["source"] == "UKB_WES"].copy()

    @property
    def hgnc_path(self) -> str:
        return os.path.join(self.phewas_dir, "gene_with_protein_product.txt")

    def resolve_gene(self) -> tuple[str, str]:
        if self._approved_symbol is None or self._ensembl_id is None:
            approved_symbol, ensembl_id, _entrez_id = get_gene_info(self.input_gene, self.hgnc_path)
            self._approved_symbol = approved_symbol
            self._ensembl_id = ensembl_id
        return self._approved_symbol, self._ensembl_id

    @property
    def approved_symbol(self) -> str:
        approved_symbol, _ensembl_id = self.resolve_gene()
        return approved_symbol

    @property
    def ensembl_id(self) -> str:
        _approved_symbol, ensembl_id = self.resolve_gene()
        return ensembl_id

    @property
    def display_gene(self) -> str:
        return self.approved_symbol

    @property
    def requested_phewas_filename(self) -> str:
        if self.phewas_filename:
            return self.phewas_filename
        return f"{self.input_gene}_genebass_phewas.csv"

    @property
    def requested_phewas_file(self) -> str:
        return os.path.join(self.phewas_dir, self.requested_phewas_filename)

    @property
    def approved_phewas_filename(self) -> str:
        return f"{self.approved_symbol}_genebass_phewas.csv"

    @property
    def approved_phewas_file(self) -> str:
        return os.path.join(self.phewas_dir, self.approved_phewas_filename)

    def existing_phewas_file(self) -> str | None:
        if os.path.exists(self.requested_phewas_file):
            return self.requested_phewas_file
        if self.requested_phewas_file != self.approved_phewas_file and os.path.exists(self.approved_phewas_file):
            return self.approved_phewas_file
        return None

    def run(self, playwright: Playwright) -> None:
        browser = playwright.chromium.launch(headless=True)
        context = browser.new_context(accept_downloads=True)
        page = context.new_page()
        query_id = f"https://app.genebass.org/gene/{self.ensembl_id}?burdenSet=pLoF&phewasOpts=1&resultLayout=full"
        page.goto(query_id)
        page.click("label:has-text(\"Burden\")")
        page.wait_for_timeout(random.uniform(100, 1000))
        with page.expect_download() as download_info:
            page.click("text=Export data to CSV")
        download = download_info.value
        os.makedirs(self.phewas_dir, exist_ok=True)
        download.save_as(path=self.approved_phewas_file)
        page.wait_for_timeout(random.uniform(100, 1000))
        context.close()
        browser.close()

    def fetch_phewas_file(self) -> str:
        with sync_playwright() as playwright:
            self.run(playwright)
        return self.approved_phewas_file

    def ensure_phewas_file(self) -> str:
        phewas_file = self.existing_phewas_file()
        if phewas_file:
            return phewas_file
        return self.fetch_phewas_file()

    @staticmethod
    def clean_phewas(phewas_file: str) -> pd.DataFrame:
        df = pd.read_csv(phewas_file, sep=",")
        return df.loc[df["Trait type"] == "icd_first_occurrence"]

    def get_manhattan(self, phewas_file: str):
        df = self.clean_phewas(phewas_file)
        p_col = "P-Value (Burden)"
        cat_col = "Category"

        df[p_col] = pd.to_numeric(df[p_col], errors="coerce")
        df = df[df[p_col].notna() & (df[p_col] > 0) & np.isfinite(df[p_col])]
        prefix = "Health-related outcomes > First occurrences > "
        df[cat_col] = df[cat_col].astype(str).str.replace(prefix, "", regex=False).fillna("Unknown")
        df[cat_col] = df[cat_col].map(_GENEBASS_CATEGORY_SHORT).fillna(df[cat_col])
        sig_p = 0.05 / len(df)

        fig, ax, plot_df = manhattan(
            df,
            p_col=p_col,
            category_col=cat_col,
            label_col="Description",
            beta_col="Beta",
            sig_p=sig_p,
            top_k_labels=5,
            title=f"{self.display_gene} PheWAS GeneBass",
        )
        return fig, plot_df

    def generate_clinical_report(self, doc):
        phewas_file = self.ensure_phewas_file()
        doc.add_heading("Phenome-wide Association Study (PheWAS) in UK Biobank cohort collected from GeneBass", level=2)
        if os.path.exists(phewas_file):
            fig, plot_df = self.get_manhattan(phewas_file)
            fig_path = uth._save_fig_to_tmp(fig, basename="genebass_phewas", dpi=300)
            plt.close(fig)
            doc.add_paragraph()
            doc.add_picture(fig_path, width=Inches(6))
            uth._add_figure_caption(doc, f"{self.display_gene} PheWAS results from GeneBass.")
            doc.add_paragraph()

            top_significant = plot_df.nsmallest(10, "P-Value (Burden)")
            columns = ["Description", "Category", "Beta", "P-Value (Burden)"]
            display_df = top_significant[columns].copy()
            display_df.columns = ["Description", "Category", "Beta", "P-value"]
            uth._add_table_to_doc(
                doc,
                display_df,
                uth._add_table_title(doc, f"Top 10 significant associations for {self.display_gene} in GeneBass"),
                ["Description", "Category", "Beta", "P-value"],
                list(map(Inches, [2.75, 2.0, 0.5, 0.75])),
            )

            negative_beta = plot_df[plot_df["Beta"] < 0].nsmallest(10, "P-Value (Burden)")
            if len(negative_beta) > 0:
                display_df_neg = negative_beta[columns].copy()
                display_df_neg.columns = ["Description", "Category", "Beta", "P-value"]
                uth._add_table_to_doc(
                    doc,
                    display_df_neg,
                    uth._add_table_title(doc, f"Top 10 protective associations for {self.display_gene} in GeneBass"),
                    ["Description", "Category", "Beta", "P-value"],
                    list(map(Inches, [2.75, 2.0, 0.5, 0.75])),
                )
            doc.add_page_break()
        else:
            doc.add_paragraph(f"No GeneBass PheWAS results found for {self.display_gene}.")
            doc.add_page_break()
        return doc

    def generate_labs_measurements_report(self, doc):
        phewas_file = self.ensure_phewas_file()
        doc.add_heading("Association study results of labs and measurements in UK Biobank cohort collected from GeneBass", level=2)
        if os.path.exists(phewas_file):
            phewas_df = pd.read_csv(phewas_file, sep=",")
            phewas_df = phewas_df.loc[phewas_df["Trait type"] == "continuous"]
            columns = ["Description", "Category", "Beta", "P-Value (Burden)"]
            display_df = phewas_df[columns].copy()
            display_df.columns = ["Phenotype", "Category", "Beta", "P-value"]
            display_df = display_df.nsmallest(50, "P-value")
            doc.add_paragraph()
            uth._add_table_to_doc(
                doc,
                display_df,
                uth._add_table_title(doc, f"Top 50 associations for {self.display_gene} in GeneBass"),
                ["Phenotype", "Category", "Beta", "P-value"],
                list(map(Inches, [2.75, 1.75, 0.75, 0.75])),
            )
            doc.add_page_break()
        else:
            doc.add_paragraph()
            doc.add_paragraph(f"No GeneBass labs and measurements results found for {self.display_gene}.")
            doc.add_page_break()
        return doc

    def generate_indication_specific_report(self, doc, keywords):
        phewas_file = self.ensure_phewas_file()
        doc.add_heading("Indication specific PheWAS results in UK Biobank cohort collected from GeneBass", level=2)
        doc.add_paragraph("The keywords searched to generate the report were: " + ", ".join(keywords))
        if os.path.exists(phewas_file):
            phewas_df = pd.read_csv(phewas_file, sep=",")
            escaped_keywords = [re.escape(k) for k in keywords if k]
            if not escaped_keywords:
                doc.add_paragraph("No valid indication keywords were provided.")
                doc.add_page_break()
                return doc
            keyword_pattern = "|".join(escaped_keywords)
            phewas_df = phewas_df.loc[phewas_df.Description.str.contains(keyword_pattern, case=False, na=False)]
            columns = ["Description", "Category", "Beta", "P-Value (Burden)"]
            display_df = phewas_df[columns].copy().sort_values("P-Value (Burden)").reset_index(drop=True)
            if len(display_df) == 0:
                doc.add_paragraph("No associations found for the specified indications. Please broaden the indication keywords and try again.")
                doc.add_page_break()
            else:
                display_df.columns = ["Description", "Category", "Beta", "P-value"]
                uth._add_table_to_doc(
                    doc,
                    display_df,
                    uth._add_table_title(doc, f"Indication-specific associations for {self.display_gene} in GeneBass"),
                    ["Description", "Category", "Beta", "P-value"],
                    list(map(Inches, [2.75, 1.75, 0.75, 0.75])),
                )
                doc.add_page_break()
        else:
            doc.add_paragraph(f"No GeneBass PheWAS results found for {self.display_gene}.")
            doc.add_page_break()
        return doc

    def get_carrier_info(self) -> dict:
        normalized_gene = self.input_gene.strip().upper()
        df = self.load_plof_summary()
        hits = df.loc[df["Gene"] == normalized_gene]

        if hits.empty:
            nhomz = None
            nhetz = None
        else:
            row = hits.iloc[0]
            nhomz = None if pd.isna(row["Homozygote_Count"]) else int(row["Homozygote_Count"])
            nhetz = None if pd.isna(row["Heterozygote_Count"]) else int(row["Heterozygote_Count"])

        return {
            "gene": normalized_gene,
            "nhetz": nhetz,
            "nhomz": nhomz,
        }


def run(playwright: Playwright, gene: str, gene_id: str, save_path: str) -> None:
    client = GeneBassClient(gene=gene, phewas_dir=save_path)
    client._approved_symbol = gene
    client._ensembl_id = gene_id
    client.run(playwright)


def clean_genebass_phewas(phewas_file):
    return GeneBassClient.clean_phewas(phewas_file)


def get_genebass_manhattan(gene, phewas_file):
    client = GeneBassClient(gene=gene, phewas_dir=os.path.dirname(phewas_file))
    client._approved_symbol = gene
    return client.get_manhattan(phewas_file)


def generate_genebass_clinical_report(doc, gene, phewas_dir, phewas_filename=""):
    return GeneBassClient(gene=gene, phewas_dir=phewas_dir, phewas_filename=phewas_filename).generate_clinical_report(doc)


def generate_genebass_labs_measurements_report(doc, gene, phewas_dir, phewas_filename=""):
    return GeneBassClient(gene=gene, phewas_dir=phewas_dir, phewas_filename=phewas_filename).generate_labs_measurements_report(doc)


def generate_genebass_indication_specific_report(doc, gene, phewas_dir, keywords, phewas_filename=""):
    return GeneBassClient(gene=gene, phewas_dir=phewas_dir, phewas_filename=phewas_filename).generate_indication_specific_report(doc, keywords)


def get_genebass_carrier_info(gene: str) -> dict:
    return GeneBassClient(gene=gene, phewas_dir="").get_carrier_info()


if __name__ == "__main__":
    from docx import Document

    doc = Document()
    gene = "TNNC1"
    phewas_dir = "/home/dbanerjee/deepro/rwe/data/test/"
    client = GeneBassClient(gene=gene, phewas_dir=phewas_dir)
    doc = client.generate_clinical_report(doc)
    doc = client.generate_labs_measurements_report(doc)
    doc = client.generate_indication_specific_report(doc, keywords=["hypertrophic_cardiomyopathy"])
    doc.save(f"/home/dbanerjee/deepro/rwe/data/test/{gene}_genebass_report.docx")
    print(get_genebass_carrier_info(gene))



import os
import re
import shutil
import zipfile
from pathlib import Path
from playwright.sync_api import Playwright, sync_playwright
from rwe.plots.clinical import manhattan
import matplotlib.pyplot as plt
import pandas as pd

import rwe.utils.helpers as uth
from docx.shared import Inches

AZN_CHAPTER_SHORT = {
    "Chapter I Certain infectious and parasitic diseases": "Infectious",
    "Chapter II Neoplasms": "Neoplasms",
    "Chapter III Diseases of the blood and blood-forming organs and certain disorders involving the immune mechanism": "Blood / Immune",
    "Chapter IV Endocrine nutritional and metabolic diseases": "Endocrine / Metabolic",
    "Chapter V Mental and behavioural disorders": "Mental / Behavioral",
    "Chapter VI Diseases of the nervous system": "Nervous System",
    "Chapter VII Diseases of the eye and adnexa": "Eye",
    "Chapter VIII Diseases of the ear and mastoid process": "Ear / Mastoid",
    "Chapter IX Diseases of the circulatory system": "Circulatory",
    "Chapter X Diseases of the respiratory system": "Respiratory",
    "Chapter XI Diseases of the digestive system": "Digestive",
    "Chapter XII Diseases of the skin and subcutaneous tissue": "Skin",
    "Chapter XIII Diseases of the musculoskeletal system and connective tissue": "Musculoskeletal",
    "Chapter XIV Diseases of the genitourinary system": "Genitourinary",
    "Chapter XV Pregnancy childbirth and the puerperium": "Pregnancy / Childbirth",
    "Chapter XVII Congenital malformations deformations and chromosomal abnormalities": "Congenital",
    "Chapter XVIII Symptoms signs and abnormal clinical and laboratory findings not elsewhere classified": "Symptoms / Findings",
    "Chapter XXI Factors influencing health status and contact with health services": "Health / Services",
}

def extract_only_csv_rename(zip_path: str, gene: str, trait_type: str) -> str:
    zip_path = Path(zip_path)
    gene_dir = zip_path.parent
    target_csv = gene_dir / f"{gene}_azn_{trait_type}.csv"

    # Extract into a dedicated folder
    extract_dir = gene_dir / f"__extract_{zip_path.stem}"
    if extract_dir.exists():
        shutil.rmtree(extract_dir)
    extract_dir.mkdir(parents=True, exist_ok=True)

    with zipfile.ZipFile(zip_path, "r") as z:
        z.extractall(extract_dir)

    csvs = list(extract_dir.rglob("*.csv"))
    if not csvs:
        raise FileNotFoundError(f"No CSV found after extracting {zip_path}")

    csvs.sort(key=lambda p: p.stat().st_size, reverse=True)
    chosen = csvs[0]

    if target_csv.exists():
        target_csv.unlink()
    chosen.replace(target_csv)

    # Cleanup only what we created
    shutil.rmtree(extract_dir)
    if zip_path.exists():
        zip_path.unlink()

    return str(target_csv)


def run(playwright: Playwright, gene: str, save_path: str, trait_type="binary") -> None:
    browser = playwright.chromium.launch(headless=True)
    context = browser.new_context(accept_downloads=True)
    page = context.new_page()

    url = f"https://azphewas.com/geneView/6319c068-fd59-46d8-85ee-82d82482eb14/{gene}/glr/{trait_type}"
    page.goto(url, wait_until="networkidle")

    export_control = page.locator('[data-testid="export-settings-control"] button:has-text("Export")')
    export_control.click()

    inner_export_btn = page.get_by_test_id("export_button")
    inner_export_btn.wait_for(state="visible")

    plot_checkbox = page.locator('[data-testid="plot-export-type-toggler"] input[type="checkbox"]')
    plot_checkbox.set_checked(False, force=True)
    page.wait_for_timeout(300)

    os.makedirs(save_path, exist_ok=True)

    with page.expect_download(timeout=120_000) as dl_info:
        inner_export_btn.click(force=True)

    download = dl_info.value
    fname = download.suggested_filename
    out_dir = save_path
    os.makedirs(out_dir, exist_ok=True)
    out_path = os.path.join(out_dir, fname)
    download.save_as(out_path)
    print("Saved zip:", out_path)

    final_csv = extract_only_csv_rename(out_path, gene=gene, trait_type=trait_type)
    print("Final CSV:", final_csv)

    context.close()
    browser.close()
    return

def clean_azn_phewas(phewas_file):
    df = pd.read_csv(phewas_file, sep=",")
    df = df.loc[df["Collapsing model"]=="ptv"]
    return df

def get_azn_manhattan(gene, phewas_file):
    df = clean_azn_phewas(phewas_file)
    name_col = "Phenotype"
    p_col    = "P value"
    odds_ratio_col = "Odds ratio"
    cat_col  = "Phenotypic category"

    df[name_col] = df[name_col].astype(str).str.rsplit("#", n=1).str[-1]
    df[cat_col] = df[cat_col].map(AZN_CHAPTER_SHORT).fillna("Other")

    sig_p = 0.05/len(df)
    fig, ax, plot_df = manhattan(
        df,
        p_col=p_col,
        category_col=cat_col,
        label_col=name_col,
        odds_ratio_col=odds_ratio_col,
        sig_p=sig_p,
        top_k_labels=5,
        title=f"{gene} PheWAS AstraZeneca",
    )
    return fig, plot_df

def generate_azn_clinical_report(doc, gene, phewas_dir, phewas_filename=""):
    if not phewas_filename:
        phewas_filename = f"{gene}_azn_binary.csv"
    phewas_file = os.path.join(phewas_dir, phewas_filename)

    if not os.path.exists(phewas_file):
        with sync_playwright() as playwright:
            run(playwright, gene=gene, save_path=phewas_dir, trait_type="binary")

    doc.add_heading("Phenome-wide Association Study (PheWAS) in UK Biobank cohort collected from AstraZeneca", level=2)
    if os.path.exists(phewas_file):
        fig, plot_df = get_azn_manhattan(gene, phewas_file)
        fig_path = uth._save_fig_to_tmp(fig, basename="azn_phewas", dpi=300)
        plt.close(fig)  # close the figure to free memory
        doc.add_paragraph()  # spacing
        doc.add_picture(fig_path, width=Inches(6.5))
        uth._add_figure_caption(doc, f"{gene} PheWAS results from AstraZeneca.")
        doc.add_paragraph()  # spacing

        # Table: Top 10 significant hits (sorted by p-value)
        top_significant = plot_df.nsmallest(10, 'P value')
        columns = ['Phenotype', 'Phenotypic category', 'Odds ratio', 'P value', 'Ancestry']
        # Rename columns for display
        display_df = top_significant[columns].copy()
        display_df.columns = ['Phenotype', 'Category', 'OR', 'P-value', 'Ancestry']
        uth._add_table_to_doc(doc, display_df, uth._add_table_title(doc, f"Top 10 significant associations for {gene} in AstraZeneca"),
                        ['Phenotype', 'Category', 'OR', 'P-value', 'Ancestry'], list(map(Inches, [2.25, 1.4, 0.7, 0.85, 0.8]))
                        )
        # Table: Top 10 negative beta hits (odds_ratio < 0, sorted by p-value)
        negative_beta = plot_df[plot_df['Odds ratio'] < 1].nsmallest(10, 'P value')
        if len(negative_beta) > 0:
            display_df_neg = negative_beta[columns].copy()
            display_df_neg.columns = ['Phenotype', 'Category', 'OR', 'P-value', 'Ancestry']
            uth._add_table_to_doc(doc, display_df_neg, uth._add_table_title(doc, f"Top 10 protective associations for {gene} in AstraZeneca"),
                            ['Phenotype', 'Category', 'OR', 'P-value', 'Ancestry'], list(map(Inches, [2.25, 1.4, 0.7, 0.85, 0.8]))
                            )
        doc.add_page_break()  # spacing

    else:
        doc.add_paragraph(f"No AstraZeneca PheWAS results found for {gene}.")
        doc.add_page_break()  # spacing
    return doc


def generate_azn_labs_measurements_report(doc, gene, phewas_dir, phewas_filename=""):
    if not phewas_filename:
        phewas_filename = f"{gene}_azn_continuous.csv"
    phewas_file = os.path.join(phewas_dir, phewas_filename)

    if not os.path.exists(phewas_file):
        with sync_playwright() as playwright:
            run(playwright, gene=gene, save_path=phewas_dir, trait_type="continuous")

    doc.add_heading("Association study results of labs and measurements in UK Biobank cohort collected from AstraZeneca", level=2)
    if os.path.exists(phewas_file):
        phewas_df = clean_azn_phewas(phewas_file)
        columns = ['Phenotype', 'Phenotypic category', 'Effect size', 'P value', 'Ancestry']
        display_df = phewas_df[columns].copy()
        display_df.columns = ['Phenotype', 'Category', 'Beta', 'P-value', 'Ancestry']
        display_df = display_df.loc[display_df["Category"].isin(AZN_CHAPTER_SHORT.keys())].sort_values("P-value")
        display_df["Category"] = display_df["Category"].map(AZN_CHAPTER_SHORT).fillna(display_df["Category"])
        display_df = display_df.head(50)
        doc.add_paragraph()  # spacing
        uth._add_table_to_doc(doc, display_df, uth._add_table_title(doc, f"Top 50 associations for {gene} in AstraZeneca"),
                        ['Phenotype', 'Category', 'Beta', 'P-value', 'Ancestry'], list(map(Inches, [2.25, 1.4, 0.7, 0.85, 0.8]))
                        )
        doc.add_page_break()  # spacing
    else:
        doc.add_paragraph()  # spacing
        doc.add_paragraph(f"No AstraZeneca labs and measurements results found for {gene}.")
        doc.add_page_break()  # spacing
    return doc


def generate_azn_indication_specific_report(doc, gene, phewas_dir, keywords, trait_types=["binary", "continuous"], phewas_filenames=["", ""]):
    trait_to_effect_dict = {"binary": "Odds ratio", "continuous": "Effect size"}
    display_dfs = []
    doc.add_heading("Indication specific PheWAS results in UK Biobank cohort collected from AstraZeneca", level=2)
    doc.add_paragraph("The keywords searched to generate the report were: " + ", ".join(keywords))
    for trait, fn in zip(trait_types, phewas_filenames):
        if not fn:
            phewas_filename = f"{gene}_azn_{trait}.csv"
        else:
            phewas_filename = fn
        phewas_file = os.path.join(phewas_dir, phewas_filename)
        
        if os.path.exists(phewas_file):
            phewas_df = clean_azn_phewas(phewas_file)
            escaped_keywords = [re.escape(k) for k in keywords if k]
            if not escaped_keywords:
                continue
            keyword_pattern = "|".join(escaped_keywords)
            phewas_df = phewas_df.loc[(phewas_df.Phenotype.str.contains(keyword_pattern, case=False, na=False))]
            columns = ['Phenotype', 'Phenotypic category', trait_to_effect_dict[trait], 'P value', 'Ancestry']
            display_df = phewas_df[columns].copy()
            display_df.columns = ['Phenotype', 'Category', 'Effect', 'P-value', 'Ancestry']
            display_df["Trait type"] = trait
            display_dfs.append(display_df)
    
    if not display_dfs:
        display_df = pd.DataFrame()
    else:
        display_df = pd.concat(display_dfs, ignore_index=True).sort_values("P-value").reset_index(drop=True)
        display_df["Category"] = display_df["Category"].map(AZN_CHAPTER_SHORT).fillna(display_df["Category"])

    if len(display_df) == 0:
        doc.add_paragraph()  # spacing
        doc.add_paragraph("No associations found for the specified indications. Please broaden the indication keywords and try again.")
        doc.add_page_break()
    else:
        doc.add_paragraph()  # spacing
        uth._add_table_to_doc(doc, display_df, uth._add_table_title(doc, f"Indication-specific associations for {gene} in AstraZeneca"),
                        ['Phenotype', 'Category', 'Effect', 'P-value', 'Trait type', 'Ancestry'], list(map(Inches, [1.8, 1, 0.75, 0.75, 0.9, 0.8]))
                        )
        doc.add_page_break()
    return doc

def generate_azn_plasma_protein_biomarkers_report(doc, gene, phewas_dir, phewas_filename=""):
    if not phewas_filename:
        phewas_filename = f"{gene}_azn_continuous.csv"
    phewas_file = os.path.join(phewas_dir, phewas_filename)

    if not os.path.exists(phewas_file):
        with sync_playwright() as playwright:
            run(playwright, gene=gene, save_path=phewas_dir, trait_type="continuous")

    doc.add_heading("Association study results of plasma proteomics data in UK Biobank cohort collected from AstraZeneca", level=2)
    if os.path.exists(phewas_file):
        phewas_df = clean_azn_phewas(phewas_file)
        columns = ['Phenotype', 'Phenotypic category', 'Effect size', 'P value']
        display_df = phewas_df[columns].copy()
        display_df.columns = ['Phenotype', 'Category', 'Beta', 'P-value']
        display_df = display_df.loc[display_df["Category"]=="Proteomics"].sort_values("P-value")
        display_df = display_df.head(50)
        display_df["Phenotype"] = display_df["Phenotype"].str.split("#").str[2] + " (" + display_df["Phenotype"].str.split("#").str[3] + ")"
        doc.add_paragraph()  # spacing
        uth._add_table_to_doc(doc, display_df, uth._add_table_title(doc, f"Top 50 plasma proteomics biomarkers for {gene} in AstraZeneca"),
                        ['Phenotype', 'Category', 'Beta', 'P-value'], list(map(Inches, [2.75, 1.75, 0.75, 0.75]))
                        )
        doc.add_page_break()  # spacing
    else:
        doc.add_paragraph()  # spacing
        doc.add_paragraph(f"No AstraZeneca proteomics data found for {gene}.")
        doc.add_page_break()  # spacing
    return doc

if __name__ == "__main__":
    from docx import Document
    doc = Document()
    gene = "MBL2"
    phewas_dir = "/home/dbanerjee/deepro/rwe/data/test/"
    doc = generate_azn_clinical_report(doc, gene, phewas_dir)
    doc = generate_azn_labs_measurements_report(doc, gene, phewas_dir)
    doc = generate_azn_plasma_protein_biomarkers_report(doc, gene, phewas_dir)
    doc = generate_azn_indication_specific_report(doc, gene, phewas_dir, keywords=["diab", "obes", "kid", "ren", "fat", "weig"])
    doc.save("/home/dbanerjee/deepro/rwe/data/test/MBL2_azn_report.docx")
